#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>

#include "../nvrec.h"
#include "mcs.h"

void *mcs_accept(void * foo)
{
	mcs_t *mcs = (mcs_t *)foo;
	struct mcs_tcp *tcp = NULL;
	struct sockaddr_in addr;
	int i;
	
	do {
		if(!tcp)
			tcp = malloc(sizeof(struct mcs_tcp));
		if(!tcp) {
			PERROR("Malloc tcp");
			usleep(100000);
			continue;
		}
		tcp->sock = accept(mcs->fd, (void *)&addr, &i);
		if(tcp->sock < 0) {
			PERROR("Accept");
			usleep(100000);
			continue;
		}
		IPRINT("Got new connection\n");
		if(fcntl(tcp->sock, F_SETFL, O_NONBLOCK) != 0) {
			EPRINT("Could not set new connection to non-blocking - terminating link!\n");
			close(tcp->sock);
			continue;
		}
		pthread_mutex_lock(&mcs->tcp_lock);
		tcp->next = mcs->tcp;
		mcs->tcp = tcp;
		pthread_mutex_unlock(&mcs->tcp_lock);
		tcp = NULL;
	} while(1);
}

mcs_t *MCS_open(char * fname)
{
	struct hostent *he;
	mcs_t *mcs;
	int i;
	char *ip = NULL;
	int port;
	
	IPRINT("Opening (multi/uni)cast channel %s:\n", fname);
/* stdout */
	if(strcmp(fname, "-") == 0) {
		IPRINT(" - Playing to stdout!\n");
		mcs = (mcs_t *)malloc(sizeof(mcs_t));
		if(!mcs) {
			PERROR("malloc");
			goto error;
		}
		memset(mcs,0,sizeof(mcs_t));
		mcs->fd = -1;
		return mcs;
	}
/* tcp / udp */
	ip = fname + 6;
	i = strcspn(ip, ":");
	if(i == strlen(ip)) {
		EPRINT("Malformed UDP address <ip>:<port>\n");
		goto error;
	}
	ip[i] = 0;
	if(!(he = gethostbyname(ip))) {
		EPRINT("Could not resolve ip/hostname \"%s\".\n", ip);
		goto error;
	}
	port = atoi(ip + i + 1);
	IPRINT("IP: %s, PORT: %d\n", ip, port);
/* udp */
	if(strncmp(fname, "udp://", 6) == 0) {
		IPRINT("MULTICAST ...\n");
		if ((i = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
			PERROR("Opening socket");
			goto error;
		}
		mcs = (mcs_t *)malloc(sizeof(mcs_t));
		if(!mcs) {
			PERROR("malloc");
			close(i);
			goto error;
		}
		memset(mcs,0,sizeof(mcs_t));
		mcs->fd = i;
		mcs->addr.sin_family=he->h_addrtype;
		memcpy(&(mcs->addr.sin_addr), he->h_addr_list[0], he->h_length);
		mcs->addr.sin_port=htons(port);
		return mcs;
	}
/* tcp */
	if(strncmp(fname, "tcp://", 6) == 0) {
		struct protoent *new_protocol;
		struct sockaddr_in addr;
		sigset_t set;
		int sock;
		
		IPRINT("UNICAST ...\n");
		new_protocol = getprotobyname("tcp");
		endprotoent();
		if (!new_protocol) {
			EPRINT("Could not find TCP protocol!");
			goto error;
		}
		
		if ((sock = socket(PF_INET, SOCK_STREAM, new_protocol->p_proto)) < 0) {
			PERROR("Opening socket");
			goto error;
		}
		i = 1;
		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i));
		memset(&addr, 0, sizeof(addr));
		addr.sin_family=he->h_addrtype;
		memcpy(&(addr.sin_addr), he->h_addr_list[0], he->h_length);
		addr.sin_port = htons(port);
		if (bind(sock, (void *)&addr, sizeof(addr)) < 0) {
			PERROR("TCP bind");
			close(sock);
			goto error;
		}
		if (listen(sock, 100) < 0) {
			PERROR("TCP listen");
			close(sock);
			goto error;
		}
		
		mcs = (mcs_t *)malloc(sizeof(mcs_t));
		if(!mcs) {
			PERROR("malloc");
			close(i);
			goto error;
		}
		memset(mcs, 0, sizeof(mcs_t));
		mcs->fd = sock;
		sigemptyset(&set);
		sigaddset(&set, SIGPIPE);
		sigprocmask(SIG_BLOCK, &set, NULL);
		if(pthread_mutex_init(&mcs->tcp_lock, NULL) != 0) {
			PERROR("Mutex init");
			goto error1;
		}
		if(pthread_create(&mcs->mcs_accept, NULL, mcs_accept, mcs) != 0) {
			PERROR("Pthread create");
			goto error1;
		}
		return mcs;
	}
/* oops */
	EPRINT("Oops, unhandled protocol: %5s\n", fname);
error:
	return NULL;
error1:
	close(i);
	free(mcs);
	return NULL;
}

void MCS_set_video(mcs_t *mcs, int width, int height, double fps, char *compressor)
{
	mcs->system.video_width = width;
	mcs->system.video_height = height;
	mcs->system.video_size = width * height * 2;
	memcpy(&(mcs->system.video_format), compressor, 4);
	mcs->system.video_time = (uint64_t)((double)1e9 / fps);
}

void MCS_set_audio(mcs_t *mcs, int channels, long rate, int bits, double fps, int format)
{
	mcs->system.audio_rate = rate;
	mcs->system.audio_bits = bits;
	mcs->system.audio_channels = channels;
	mcs->system.audio_size = (uint32_t)((double)(rate * channels * (bits>>3)) / fps) * 4;
	mcs->system.audio_uncomp_size = mcs->system.audio_size * 4;
	mcs->system.audio_format = format;
	mcs->system.audio_time = (uint64_t)((double)1e9 / fps);
}

int mcs_write(mcs_t *mcs, char *data, long bytes)
{
	long i;
	struct mcs_tcp *tcp = NULL, *tcp1;
	
	while(bytes > 0) {
		i = MCS_CHUNK_SIZE - mcs->buffer_pos;
		if(i > bytes)
			i = bytes;
		memcpy(mcs->buffer + mcs->buffer_pos, data, i);
		bytes -= i;
		mcs->buffer_pos += i;
		data += i;
		if(mcs->buffer_pos == MCS_CHUNK_SIZE) {
			if(mcs->fd < 0) {
				write(STDOUT_FILENO,
					mcs->buffer,
					MCS_CHUNK_SIZE);
			} else if(mcs->mcs_accept) {
				pthread_mutex_lock(&mcs->tcp_lock);
				tcp = mcs->tcp;
				while(tcp) {
					i = write(tcp->sock, 
						mcs->buffer, 
						MCS_CHUNK_SIZE);
					if((i == -1) && (errno == EPIPE)) {
						IPRINT("closing connection\n");
						if(tcp == mcs->tcp) {
							mcs->tcp = tcp->next;
							free(tcp);
							tcp = mcs->tcp;
							continue;
						} else {
							for(tcp1=mcs->tcp; tcp1; tcp1=tcp1->next) {
								if(tcp1->next == tcp)
									break;
							}
							tcp1->next = tcp->next;
							free(tcp);
							tcp = tcp1->next;
							continue;
						}
					}
					tcp = tcp->next;
				}
				pthread_mutex_unlock(&mcs->tcp_lock);
			} else {
				sendto(mcs->fd, 
					mcs->buffer,
					MCS_CHUNK_SIZE,
					0,
					(struct sockaddr *)&mcs->addr,
					sizeof(mcs->addr));
			}
			mcs->buffer_pos = 0;
		}
	}
	return 0;
}

int mcs_system(mcs_t *mcs)
{
	if(mcs->pkt_count == 0) {
		mcs->header.sync = MCS_SYNC;
		mcs->header.time = 0;
		mcs->header.size = sizeof(mcs->system);
		mcs->header.type = MCS_FRAME_SYSTEM;
		mcs_write(mcs, (void *)&mcs->header, sizeof(mcs->header));
		mcs_write(mcs, (void *)&mcs->system, sizeof(mcs->system));
	}
	mcs->pkt_count++;
	if(mcs->pkt_count == MCS_HEADER_CNT)
		mcs->pkt_count = 0;
	return 0;
}

int  MCS_write_frame(mcs_t *mcs, char *data, long bytes, uint64_t time)
{
	mcs_system(mcs);
	mcs->header.sync = MCS_SYNC;
	mcs->header.time = time;
	mcs->header.size = bytes;
	mcs->header.type = MCS_FRAME_VIDEO;
	mcs_write(mcs, (void *)&mcs->header, sizeof(mcs->header));
	mcs_write(mcs, data, bytes);
	return 0;
}


int  MCS_write_audio(mcs_t *mcs, char *data, long bytes, uint64_t time)
{
	mcs_system(mcs);
	mcs->header.sync = MCS_SYNC;
	mcs->header.time = time;
	mcs->header.size = bytes;
	mcs->header.type = MCS_FRAME_AUDIO;
	mcs_write(mcs, (void *)&mcs->header, sizeof(mcs->header));
	mcs_write(mcs, data, bytes);
	return 0;
}

void MCS_close(mcs_t *mcs)
{
	struct mcs_tcp *tcp;
	if(mcs->mcs_accept) {
		pthread_kill(mcs->mcs_accept, SIGKILL);
		while(mcs->tcp) {
			close(mcs->tcp->sock);
			tcp = mcs->tcp->next;
			free(mcs->tcp);
			mcs->tcp = tcp;
		}
	}
	close(mcs->fd);
	free(mcs);
}
