/* 
   nvrec_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    deinterlace filters from xawtv deinterlacing plugin by:
    	Conrad Kreyling <conrad@conrad.nerdland.org>
    	Patrick Barrett <yebyen@nerdland.org>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include "tick.h"

#include "nvrec.h"
#if HAVE_X86_MMX > 0
#include "cpu_features/cpu_features.h"
#include "deint_fieldbob_mmx.h"
#include <inttypes.h>
#endif

#define ABUFFERSIZE 65536

volatile int nvcore_fin = 0;
volatile int nvcore_die = 0;
int nvcore_verbose = NV_VERBOSE_INFO;

void * nvrec_core_thread(void *t)
{
	double f = *(double *)t;
	struct sched_param sp;

	sp.sched_priority = sched_get_priority_min(SCHED_OTHER);
	pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp);

	while (!nvcore_die) {
		sync();
		usleep(1000000.0 * f);
		DPRINT("(synch)");
	}
	DXPRINT("leaving main loop\n");
	return NULL;
}

void nvcore_quit_handler(int i)
{
	DXPRINT("enter\n");
	if(!nvcore_fin) {
		IPRINT("Exiting!\n");
		IPRINT("  *** If the movie is long, this could take a while, and some OSS errors\n");
		IPRINT("  *** may be displayed.\n");
	}
	nvcore_fin = 1;
	DXPRINT("leave\n");
}

void nvcore_deint_double(unsigned char *f, int w, int h, int bpl)
{
	int i;
	
	for(i = 0; i < h; i+=2) {
		memcpy(f + bpl, f, bpl);
		f += bpl << 1;
	}
}

void nvcore_deint_bilin(unsigned char *f, int w, int h, int bpl)
{
	int x, y, t;
	
	for(y = 1; y < h - 1; y += 2) {
		for(x = 0; x < bpl; x++) {
			t = f[(y - 1) * bpl + x];
			t += f[(y + 1) * bpl + x];
			f[y * bpl + x] = t >> 1;
		}
	}
}

void nvcore_deint_cubic(unsigned char *f, int w, int h, int bpl)
{
	int x, y, t;
	
	for(y = 3; y < h - 3; y += 2) {
		for(x = 0; x < bpl; x++) {
			t = -(int)(f[(y - 3) * bpl + x]);
			t += 9 * (int)(f[(y - 1) * bpl + x]);
			t += 9 * (int)(f[(y + 1) * bpl + x]);
			t -= (int)(f[(y + 3) * bpl + x]);
			f[y * bpl + x] = t >> 4;
		}
	}
}

#if HAVE_X86_MMX > 0
/*
   DeinterlaceFieldBob algorithm
   Stolen shamelessly from an adaption for transcode by Miguel Freitas
   Based on Virtual Dub plugin by Gunnar Thalin
   MMX asm version from dscaler project (deinterlace.sourceforge.net)
   Linux version for Xine player by Miguel Freitas
   Todo: use a MMX optimized memcpy
*/
void nvcore_deint_bob(unsigned char *f, int w, int h, int bpl)
{
  uint8_t *psrc = f;
  uint8_t *pdst = f; // perhaps a bad idea... but seems to work fine [transcode has seperate in/out buffers for filters]
  int width = w; int height = h;

  int Line;
  long long* YVal1;
  long long* YVal2;
  long long* YVal3;
  long long* Dest;
  uint8_t* pEvenLines = psrc;
  uint8_t* pOddLines = psrc+width;
  int LineLength = width;
  int Pitch = width * 2;
  int IsOdd = 1;
  long EdgeDetect = 625;
  long JaggieThreshold = 73;

  int n;

  unsigned long long qwEdgeDetect;
  unsigned long long qwThreshold;
  const unsigned long long Mask = 0xfefefefefefefefe;
  const unsigned long long YMask = 0x00ff00ff00ff00ff;

  qwEdgeDetect = EdgeDetect;
  qwEdgeDetect += (qwEdgeDetect << 48) + (qwEdgeDetect << 32) + (qwEdgeDetect << 16);
  qwThreshold = JaggieThreshold;
  qwThreshold += (qwThreshold << 48) + (qwThreshold << 32) + (qwThreshold << 16);


  // copy first even line no matter what, and the first odd line if we're
  // processing an odd field.
  memcpy(pdst, pEvenLines, LineLength);
  if (IsOdd)
    memcpy(pdst + LineLength, pOddLines, LineLength);

  height = height / 2;
  for (Line = 0; Line < height - 1; ++Line)
  {
    if (IsOdd)
    {
      YVal1 = (long long *)(pOddLines + Line * Pitch);
      YVal2 = (long long *)(pEvenLines + (Line + 1) * Pitch);
      YVal3 = (long long *)(pOddLines + (Line + 1) * Pitch);
      Dest = (long long *)(pdst + (Line * 2 + 2) * LineLength);
    }
    else
    {
      YVal1 = (long long *)(pEvenLines + Line * Pitch);
      YVal2 = (long long *)(pOddLines + Line * Pitch);
      YVal3 = (long long *)(pEvenLines + (Line + 1) * Pitch);
      Dest = (long long *)(pdst + (Line * 2 + 1) * LineLength);
    }

    // For ease of reading, the comments below assume that we're operating on an odd
    // field (i.e., that bIsOdd is true).  The exact same processing is done when we
    // operate on an even field, but the roles of the odd and even fields are reversed.
    // It's just too cumbersome to explain the algorithm in terms of "the next odd
    // line if we're doing an odd field, or the next even line if we're doing an
    // even field" etc.  So wherever you see "odd" or "even" below, keep in mind that
    // half the time this function is called, those words' meanings will invert.

    // Copy the odd line to the overlay verbatim.
    memcpy((char *)Dest + LineLength, YVal3, LineLength);

    n = LineLength >> 3;
    while( n-- )
    {
      movq_m2r (*YVal1++, mm0);
      movq_m2r (*YVal2++, mm1);
      movq_m2r (*YVal3++, mm2);

      // get intensities in mm3 - 4
      movq_r2r ( mm0, mm3 );
      movq_r2r ( mm1, mm4 );
      movq_r2r ( mm2, mm5 );

      pand_m2r ( *&YMask, mm3 );
      pand_m2r ( *&YMask, mm4 );
      pand_m2r ( *&YMask, mm5 );

      // get average in mm0
      pand_m2r ( *&Mask, mm0 );
      pand_m2r ( *&Mask, mm2 );
      psrlw_i2r ( 01, mm0 );
      psrlw_i2r ( 01, mm2 );
      paddw_r2r ( mm2, mm0 );

      // work out (O1 - E) * (O2 - E) / 2 - EdgeDetect * (O1 - O2) ^ 2 >> 12
      // result will be in mm6

      psrlw_i2r ( 01, mm3 );
      psrlw_i2r ( 01, mm4 );
      psrlw_i2r ( 01, mm5 );

      movq_r2r ( mm3, mm6 );
      psubw_r2r ( mm4, mm6 );	//mm6 = O1 - E

      movq_r2r ( mm5, mm7 );
      psubw_r2r ( mm4, mm7 );	//mm7 = O2 - E

      pmullw_r2r ( mm7, mm6 );		// mm6 = (O1 - E) * (O2 - E)

      movq_r2r ( mm3, mm7 );
      psubw_r2r ( mm5, mm7 );		// mm7 = (O1 - O2)
      pmullw_r2r ( mm7, mm7 );	// mm7 = (O1 - O2) ^ 2
      psrlw_i2r ( 12, mm7 );		// mm7 = (O1 - O2) ^ 2 >> 12
      pmullw_m2r ( *&qwEdgeDetect, mm7 );// mm7  = EdgeDetect * (O1 - O2) ^ 2 >> 12

      psubw_r2r ( mm7, mm6 );      // mm6 is what we want

      pcmpgtw_m2r ( *&qwThreshold, mm6 );

      movq_r2r ( mm6, mm7 );

      pand_r2r ( mm6, mm0 );

      pandn_r2r ( mm1, mm7 );

      por_r2r ( mm0, mm7 );

      movq_r2m ( mm7, *Dest++ );
    }
  }

  // Copy last odd line if we're processing an even field.
  if (! IsOdd)
  {
    memcpy(pdst + (height * 2 - 1) * LineLength,
                      pOddLines + (height - 1) * Pitch,
                      LineLength);
  }

  // clear out the MMX registers ready for doing floating point
  // again
  emms();

}
#endif

struct s2 {
	unsigned short lo;
	unsigned short hi;
};

typedef union {
	unsigned int ui;
	struct s2 us;
} ws_t;

int nvcore_main(char **args, struct cap_core_t * cct, struct aud_core_t * audt, struct file_core_t * fct, ntime_t tofs, unsigned int vfmt)
{
	ntime_t atime, vtime, fvtime = 0;
	double capstat, fatime = 0, ftemp = 0, lastoff = 0;
	int i = 0, drate = 0, cseq = 0;
	unsigned char abuffer[ABUFFERSIZE];
	unsigned char **vframe;
	unsigned long long fdrop = 0;
	signed long long fcount = -1;
	ws_t dseq;
	pthread_t sthread;
	double Stime = 0, Qlow = 10.0, Qfast = 20.0, Qhi = 40.0, Qscale;
	struct sigaction newaction;
	int w = -1;
	int h = -1;
	void (*deint)(unsigned char *f, int w, int h, int bpl) = NULL;
	unsigned long long vrate = 0;
	ticker_t vrate_ticker;
	int vn;
	NV_ARGS;
	
#if HAVE_X86_MMX > 0
	cpu_features_init();
#endif
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-L") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -L.\n");
				goto usage;
			}
			Qlow = atof(NV_THIS_ARG);
			if((Qlow < 0) || (Qlow > 100)) {
				EPRINT("Invalid argument to -L (0..100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-M") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -M.\n");
				goto usage;
			}
			Qfast = atof(NV_THIS_ARG);
			if((Qfast < 0) || (Qfast > 100)) {
				EPRINT("Invalid argument to -M (0..100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-H") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -H.\n");
				goto usage;
			}
			Qhi = atof(NV_THIS_ARG);
			if((Qhi < 0) || (Qhi > 100)) {
				EPRINT("Invalid argument to -H (0..100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-S") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -S.\n");
				goto usage;
			}
			Stime = atof(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-F") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -F.\n");
				goto usage;
			}
			fcount = atoll(NV_THIS_ARG);
			if(fcount <= 0) {
				EPRINT("Invalid argument to -F (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			w = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			h = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-deint") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -deint.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "none") == 0) {
				deint = NULL;
				IPRINT("Deinterlace method = none\n");
			} else if(strcmp(NV_THIS_ARG, "double") == 0) {
				deint = nvcore_deint_double;
				IPRINT("Deinterlace method = double\n");
			} else if(strcmp(NV_THIS_ARG, "bilin") == 0) {
				deint = nvcore_deint_bilin;
				IPRINT("Deinterlace method = bilinear\n");
			} else if(strcmp(NV_THIS_ARG, "cubic") == 0) {
				deint = nvcore_deint_cubic;
				IPRINT("Deinterlace method = cubic\n");
#if HAVE_X86_MMX > 0
			} else if(strcmp(NV_THIS_ARG, "bob") == 0) {
				if(!CPU_HAS_FEATURE_MMX) {
					EPRINT("deint=bob but MMX not supported!\n");
					goto usage;
				}
				deint = nvcore_deint_bob;
				IPRINT("Deinterlace method = bob\n");
#endif
			} else {
				EPRINT("Unknown deinterlace method %s\n", NV_THIS_ARG);
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
	}
	
	if(!vrate)
		vrate = cct->cap_core_ftime(cct);
	nvrec_ticker_init(&vrate_ticker, vrate, cct->cap_core_ftime(cct));

	Qscale = 65536.0 / (100.0 - Qhi);
	sync();
	if(Stime > 0) {
		IPRINT("Using separate sync() thread\n");
		if(pthread_create(&sthread, NULL, nvrec_core_thread, (void *)&Stime) != 0)
			EPRINT("Could not start sync thread: continuing anyway.\n");
	} else if(Stime < 0) {
		IPRINT("Using synchronous writes\n");
	} else {
		IPRINT("Not using any sync() method\n");
	}

	sigfillset(&newaction.sa_mask);
	newaction.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	newaction.sa_handler = nvcore_quit_handler;
	if(sigaction(SIGINT, &newaction, NULL) < 0) {
		PERROR("nvrec: could not set sigint handler");
		return -1;
	}
	if(sigaction(SIGPIPE, &newaction, NULL) < 0) {
		PERROR("nvrec: could not set sigpipe handler");
		return -1;
	}
	
	dseq.ui = 0;
	
	/* Get initial synch */
	atime = audt->aud_core_tailtime(audt);
	cct->cap_core_next(cct, &vtime);
	vtime -= tofs;
	if(atime > vtime) {
		do {
			cct->cap_core_next(cct, &vtime);
			vtime -= tofs;
		}  while (atime > vtime);
	} else {
		do {
			audt->aud_core_droptail(audt);
			atime = audt->aud_core_tailtime(audt);
		}  while (atime < vtime);
	}
	if(tofs > 0) {
		vtime += tofs;
		while(atime < vtime) {
			i = audt->aud_core_read(audt, vtime, 0, abuffer, ABUFFERSIZE);
			fct->file_core_audio(fct, abuffer, i);
			atime = audt->aud_core_tailtime(audt);
		}
	}
	IPRINT("Initial sync - audio: %lli video %lli (delta %f secs)\n", atime, vtime, (atime-vtime)/1e9);

	/* Now loop until we're done! */
	while((!nvcore_fin) && (fcount)) {
		vn = nvrec_ticker_tick(&vrate_ticker);
		if(!(vframe = cct->cap_core_next(cct, &vtime))) {
			IPRINT("Error getting more video data. Trying to continue.\n");
			continue;
		}
		capstat = cct->cap_core_status(cct);
		while(vn--) {
			if(capstat > Qhi) {
				dseq.ui += (int)((capstat - Qhi) * Qscale);
			} else if(capstat < Qlow) {
				usleep(cct->cap_core_ftime(cct)/1000);
			}
			if(dseq.us.hi) {
				dseq.us.hi = 0;
				fct->file_core_nullvideo(fct);
				fdrop++;
			} else {
				if(capstat > Qfast) {
					fct->file_core_fastvideo(fct, vframe);
				} else {
					if(deint) {
						switch(vfmt) {
							case NV_YUV420:
								deint(vframe[0], w, h, w);
								deint(vframe[1], w/2, h/2, w/2);
								deint(vframe[2], w/2, h/2, w/2);
								break;
							case NV_RGB24:
								deint(vframe[0], w, h, w*3);
								break;
							case NV_YUV422:
								deint(vframe[0], w, h, w*2);
								break;
						}
					}
					fct->file_core_video(fct, vframe);
				}
			}
			i = audt->aud_core_read(audt, vtime, drate, abuffer, ABUFFERSIZE);
			fct->file_core_audio(fct, abuffer, i);
			
			if(Stime < 0)
				fct->file_core_sync(fct);
	
			fvtime += vrate;
			ftemp = (double)i;
			ftemp *= 1e9;
			ftemp /= (double)audt->aud_core_bps(audt); // in nsecs
			fatime += ftemp;
		}
		atime = audt->aud_core_tailtime(audt);

		/* Do this approx once per second.  The actual timing is not critical. */
		/* This is a clasic PI controller.  The input is the diffence between the */
		/* audio and video time. The output is the variable drate, which is passed */
		/* to the oss core to control the stretching.  drate varies between -1000 */
		/* (double speed) and 1000 (half speed) */
		if ( (cseq = (cseq+1) % 30) == 0) {
			/* total nsec diff between A and V */
			ftemp = fatime - (double)fvtime;
			/* This is the P component. 1e7 gives 100 for every second of drift */ 
			drate = (int)(- ftemp / 1e7);
			/* This is the I component. lastoff is the accumulated error. It is */
			/* required to eliminate systematic offset.  Must be smaller than P */
			/* component for stability. */
			drate -= (int)(lastoff / 1e8);
			lastoff = lastoff + ftemp;
			if(drate >= 1000) drate = 999;
			if(drate <= -1000) drate = -999;
		}
		IPRINT("VQ:% 8.03f Ofs: % 8.03f Slip: % 8.03f Warp: %d Drop: %lld   \r",
			capstat, (atime-vtime)/1e9, ftemp / 1e9, drate, fdrop);
		if(fcount > 0) fcount--;
	}
	DXPRINT("capture loop done\n");
	nvcore_die = 1;
	DXPRINT("leave\n");
	return 0;

usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	return -1;
}

void nvcore_sigrestore()
{
	struct sigaction newaction;
	DXPRINT("enter\n");
	memset(&newaction, 0, sizeof(newaction));
	newaction.sa_handler = SIG_DFL;
	DXPRINT("restroring sigint\n");
	if(sigaction(SIGINT, &newaction, NULL) < 0)
		PERROR("nvrec: could not set sigint handler");
	DXPRINT("restroring sigpipe\n");
	if(sigaction(SIGPIPE, &newaction, NULL) < 0)
		PERROR("nvrec: could not set sigint handler");
	DXPRINT("leave\n");
}

void nvcore_usage()
{
	MPRINT("NVREC core usage:\n");
	MPRINT("\t-H <%% buffer usage before dropping>  [40]\n");
	MPRINT("\t-L <min %% buffer usage>              [10]\n");
	MPRINT("\t-M <%% buffer usage before fast comp> [20]\n");
	MPRINT("\t-F <number of frames to grab>        [inf]\n");
	MPRINT("\t-S <seconds between syncs>           [1.0]\n");
	MPRINT("\t-C <config file>                     [none]\n");
#if HAVE_X86_MMX > 0
	MPRINT("\t-deint <none|double|bilin|cubic|bob> [none]\n");
#else
	MPRINT("\t-deint <none|double|bilin|cubic>     [none]\n");
#endif
	MPRINT("\t-vr <float: desired frame rate>      [src rate]\n");
	MPRINT("\t-D enables debug messages            [disabled]\n");
	MPRINT("\t-DX enables debugging of exit        [disabled]\n");
	MPRINT("\t-Q quiet operation                   [info]\n");
}

/*
	Parse the arguments. Define defaults, and set overrides.
	All modules parse arguments left to right, so we set the defaults
	first, then the user arguments, then the overrides, then the filename.
*/

char *nvcore_defaults[] = {"-d", "/dev/dsp",
			   "-r", "44100",
			   "-b", "16",
			   "-v", "/dev/video",
			   "-w", "384",
			   "-h", "288",
			   "-N", "60",
			   "-L", "10",
			   "-H", "40",
			   "-M", "20",
			   "-m",
			   NULL};

void nvcore_add_arg(char **args, char *newarg)
{
	int i = 1; /* start at one, so we always have a NULL marker at the end! */
	while((*args) && (i < NV_MAX_ARGS)) {
		args++;
		i++;
	}
	if(i == NV_MAX_ARGS) {
		EPRINT("Out of argument space, please report this bug!\n");
		return; 
	}
	*args = strdup(newarg);
	if(!(*args)) {
		PERROR("Could not malloc new arg!\n");
		return; 
	}
	return;
}

#define NV_MAX_ARG_LENGTH 256
void nvcore_parse_file(char *fname, char* prog, char **args){
	FILE *f;
	char arg[NV_MAX_ARG_LENGTH];
	char identifier[NV_MAX_ARG_LENGTH];
	int state = 0;
	int c, i = 0;

	if(!(f = fopen(fname, "r"))) {
		EPRINT("Error opening \"%s\"\n", fname);
		return;
	}
	memset(arg, 0, NV_MAX_ARG_LENGTH);
	memset(identifier, 0, NV_MAX_ARG_LENGTH);
	while((c = fgetc(f)) != EOF) {
		switch(state) {
			case 0: /* Reading command */
				if((c == ' ') || (c == '\n') || (c == '\r') || (c == '\t') || (c == '#') || (c == '{')) {
					if(c == '{'){
						state = 2;
					}
					if(strlen(arg)) {
						nvcore_add_arg(args, arg);
						memset(arg, 0, NV_MAX_ARG_LENGTH);
						i = 0;
					}
					if(c == '#') {
						state = 1;
					}
				} else {
					arg[i++] = c;
				}
				break;
			case 1: /* Comment */
				if(c == '\n') {
					state = 0;
				}
				break;
			case 2: /* looking for identifier */
				if((c == ' ') || (c == '\n') || (c == '\r') || (c == '\t') || (c == '#')) {
					if(strlen(arg)) {
						if(!strcmp(prog,arg)){
							strcpy(identifier,arg);
						}
						else if(strlen(identifier)){
							nvcore_add_arg(args, arg);
						}
						memset(arg, 0, NV_MAX_ARG_LENGTH);
						i = 0;
					}
				}
				else if(c == '}'){
					memset(identifier, 0, NV_MAX_ARG_LENGTH);
					if(strlen(arg)) {
						EPRINT("%s\n",arg);
					}
					state = 1;
				}
				else{
					arg[i++] = c;
				}
				break;
		}
	}
	if(strlen(arg)){
		nvcore_add_arg(args, arg);
	}
	fclose(f);
}

char **nvcore_args(int argc, char **argv, struct cap_core_t * cct, struct aud_core_t * audt, struct file_core_t * fct, char* prog){
	int i = 0;
	char ** args;
	char ** arg = NULL;
	char * home;
	char home_conf[256];
	
	args = (char **)malloc(sizeof(char *) * NV_MAX_ARGS);
	if(!args) {
		PERROR("nvcore args: could not allocate memory!");
		return NULL;
	}
	memset(args, 0, sizeof(char *) * NV_MAX_ARGS);
	nvcore_add_arg(args, argv[0]);
	for(arg = nvcore_defaults; *arg; arg++){
		nvcore_add_arg(args, *arg);
	}
	nvcore_parse_file(NV_CONF_FILE, prog, args);
	home = getenv("HOME");
	if(!home) {
		EPRINT("Could not find home directory!\n");
	} else {
		snprintf(home_conf, 256, "%s%s", home, NV_HOME_CONF);
		nvcore_parse_file(home_conf, prog, args);
	}
	for(i = 1; i < argc; i++) {
		if(strcmp(argv[i], "-C") == 0) {
			if((argc - i) > 1) {
				nvcore_parse_file(argv[i+1], prog, args);
				i++;
			} else {
				EPRINT("Error parsing additional config file!\n");
			}
		} else {
			nvcore_add_arg(args, argv[i]);
		}
	}
	if(cct->cap_core_overrides)
		cct->cap_core_overrides(args);
	if(audt->aud_core_overrides)
		audt->aud_core_overrides(args);
	if(fct->file_core_overrides)
		fct->file_core_overrides(args);
	for(arg = args; *arg; arg++) {
		if(strcmp(*arg, "-Q") == 0) {
			nvcore_verbose = NV_VERBOSE_QUIET;
		} else if(strcmp(*arg, "-D") == 0) {
			nvcore_verbose = NV_VERBOSE_DEBUG;
		} else if(strcmp(*arg, "-DX") == 0) {
			nvcore_verbose = NV_VERBOSE_DEBUG_EXIT;
		} 
	}
	return args;
}

void nvcore_close(char ** args)
{
	int i = 1;
	char **argp = args;
	DXPRINT("enter\n");
	while((*argp) && (i < NV_MAX_ARGS)) {
		DPRINT(" arg %i: %s\n", i, *argp);
		free(*argp);
		argp++;
		i++;
	}
	DXPRINT("free'ed individual args\n");
	free(args);
	DXPRINT("leave\n");
}
