/* 
   oss_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <pthread.h>
#include <signal.h>

#include "nvrec.h"

struct frag_t {
	unsigned char *buf;	// data buffer
	ntime_t time;		// timestamp
	int full;		// buffer full
};

struct oss_core_t {
	int fd;			// file descriptor
	
	int rate;		// image width we got
	int bits;		// image height we got
	int chans;		// pixel format we got
	
	int bps;		// bytes per second
	int realbps;		// bytes per second
	int fragsize;		// size of each fragment
	int nfrags;		// number of fragments in buffer
	
	struct frag_t *frag;	// audio buffer fragments
	int head;		// buffer head
	int tail;		// buffer tail
	int exit;		// stop audio thread
	int done;		// exited succesfully
};

void * oss_core_thread(void * t)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)t;
	int i, n;
	struct audio_buf_info info;
	ntime_t ntime;
	unsigned long long u;
	struct sched_param sp;
	sigset_t newmask;
	
	sigfillset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, NULL) != 0) {
		PERROR("Could not block signals, but continuing anyway");
	}
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				PERROR("Could not get any realtime priority");
				EPRINT("\t- This is not a serious problem, but capture quality under high\n");
				EPRINT("\t  load could be improved by running as root.\n");
			}
		}
	}

	IPRINT("Starting!\n");
	
	while(!osst->exit) {
		if ((osst->frag + osst->head)->full == 1) {
			EPRINT("Buffer full - discarding data\n");
			usleep(10000);
			continue;
		}
		i = 0;
		while(i < osst->fragsize) {
			n = read(osst->fd, (osst->frag + osst->head)->buf + i,  osst->fragsize - i);
			i += n;
		}
		ntime = getntime();
		if (ioctl(osst->fd, SNDCTL_DSP_GETISPACE, &info) < 0) {
			PERROR("Could not determine buffer offset");
			u = 0;
		} else {
			u = info.bytes;
			u *= 1000000000ULL;
			u /= osst->realbps;
		}
		i = osst->head;
		i++;
		if(i >= osst->nfrags) i = 0;
		(osst->frag + osst->head)->full = 1;
		(osst->frag + osst->head)->time = ntime - u;
		osst->head = i;
	}
	DXPRINT("main loop done\n");
	IPRINT("Exiting!\n");
	DXPRINT("signal main loop done\n");
	osst->done = 1;
	DXPRINT("leave\n");
	return NULL;
}


void oss_core_freefrags(volatile struct oss_core_t * osst)
{
	DXPRINT("enter\n");
	if(!osst->frag)
		return;
	DXPRINT("free loop\n");
	while(osst->nfrags > 0) { // FIXME - why does this segfault?
		osst->nfrags--;
/*		IPRINT("oss core close - free buffer %d (%d %d).\n", osst->nfrags, osst->head, osst->tail); */
		if((osst->frag + osst->nfrags)->buf)
			free((osst->frag + osst->nfrags)->buf);
	}
	DXPRINT("free frag\n");
	free(osst->frag);
	DXPRINT("leave\n");
}

void oss_core_close(struct aud_core_t * audt)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)audt->priv;
	
	DXPRINT("enter\n");
	if(!osst)
		return;
	DXPRINT("signal oss core exit\n");
	osst->exit = 1;
	DXPRINT("wait for oss core exit\n");
	while(!osst->done) // FIXME - time this out!
		usleep(100000);
	DXPRINT("oss core exited\n");
	IPRINT("Read thread terminated.\n");
	DXPRINT("close OSS fd\n");
	close(osst->fd);
	IPRINT("Freed buffers.\n");
	DXPRINT("free oss frags\n");
	oss_core_freefrags(osst);
	IPRINT("Freed fragments.\n");
	DXPRINT("free priv\n");
	free(audt->priv);
	DXPRINT("free audt\n");
	free(audt);
	DXPRINT("leave\n");
	return;
}

ntime_t oss_core_tailtime(struct aud_core_t * audt)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)audt->priv;

	if(!osst)
		return 0;
	while(osst->frag[osst->tail].full == 0) {
		usleep(10000);
	}

	return (osst->frag + osst->tail)->time;
}

inline void oss_core_droptail(struct aud_core_t * audt)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)audt->priv;
	int i;
	

	if(!osst)
		return;
	i = osst->tail;
	(osst->frag + i)->full = 0;
	i ++;
	if(i >= osst->nfrags) i = 0;
	osst->tail = i;
}

//FIXME: will work for little-endian only... (SLD)
struct s2 {
	unsigned short lo;
	unsigned short hi;
};
typedef union {
	unsigned int ui;
	struct s2 us;
} ws_t;

int oss_core_read(struct aud_core_t * audt, ntime_t t, int drate, unsigned char * buf, int n)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)audt->priv;
	int i, k;
	int c = 0;
	unsigned int dt = 0;
	ws_t j;
	

	if(!osst)
		return 0;
	if(drate > 0) {
		dt = drate << 15;
		dt /= 1000;  // 0 .. 32768
		dt = 65536 - dt;
	} else if(drate < 0) {
		dt = (-drate) << 15;
		dt /= 1000;  // 0 .. 32768
		dt = 65536 - dt;
		dt = (unsigned int)((unsigned long long)(1ULL<<32) / (unsigned long long)dt);
	} else {
		dt = 65536;
	}
	j.ui = 0;
	while(1) {
		i = osst->tail;
		if((osst->frag + i)->full == 0) {
			EPRINT("Ran out of audio data... not supposed to happen\n");
			return c;
		}
		if((osst->frag + i)->time > t)
			return c;
		if((n - c) < (2 * osst->fragsize))
			return c;

		if(drate == 0) {
			memcpy(buf + c, (osst->frag + i)->buf, osst->fragsize);
			c += osst->fragsize;
		} else if(osst->chans == 2) {
			unsigned int * in = (unsigned int *) ((osst->frag + i)->buf);
			unsigned int * out = (unsigned int *) (buf + c);
			k = 0;
			//SLD: Slight tweak here: retain fractional part of j between
			//fragments.  Should help even out resampling and get closer
			//to the desired number of output samples
			for(j.us.hi = 0; j.us.hi < (osst->fragsize >> 2); j.ui += dt)
				out[k++] = in[j.us.hi];
			c += k << 2;
		} else {
			unsigned short * in = (unsigned short *) ((osst->frag + i)->buf);
			unsigned short * out = (unsigned short *) (buf + c);
			k = 0;
			//SLD: Comment as above
			for(j.us.hi = 0; j.us.hi < (osst->fragsize >> 1); j.ui += dt)
				out[k++] = in[j.us.hi];
			c += k << 1;
		}
		(osst->frag + i)->full = 0;
		i ++;
		if(i >= osst->nfrags) i = 0;
		osst->tail = i;
	}
	return -1;
}

unsigned long long oss_core_bps(struct aud_core_t * audt)
{
	volatile struct oss_core_t * osst = (struct oss_core_t *)audt->priv;


	if(!osst)
		return 0;

	return osst->bps;
}

void oss_core_usage()
{
	MPRINT("OSS capture core usage:\n");
	MPRINT("\t-d <dsp device>                      [/dev/dsp]\n");
	MPRINT("\t-r <audio sample rate>               [44100]\n");
	MPRINT("\t-b <bits per sample>                 [16]\n");
	MPRINT("\t-s stereo                            [mono]\n");
	MPRINT("\t-mixsrc <dev name>:<chan name>       [none]\n");
	MPRINT("\t-mixvol <dev name>:<chan name>:<vol> [none]\n");
}

struct aud_core_t oss_core;

static char *oss_core_devnames[] = SOUND_DEVICE_NAMES;
#define MAX_DEV (sizeof(oss_core_devnames)/sizeof(char *))

int oss_core_setmixsrc(char *cmd)
{
	int fd;
	char *cn;
	int i;
	
	i = strcspn(cmd, ":");
	if(i == strlen(cmd)) {
		EPRINT("Invalid mixsrc command\n");
		EPRINT("   eg: -mixsrc /dev/mixer:line\n");
		return -1;
	}
	cmd[i] = 0;
	cn = cmd + i + 1;
	for(i = 0; i < MAX_DEV; i++) {
		if(strcmp(cn, oss_core_devnames[i]) == 0)
			break;
	}
	if(i == MAX_DEV) {
		EPRINT("Could not find mixer device \"%s\".\n", cn);
		return -1;
	}
	if ((fd = open(cmd, O_RDWR)) < 0) {
		PERROR("Could not open mixer device");
		return -1;
	}
	i = 1 << i;
	if(ioctl(fd, SOUND_MIXER_WRITE_RECSRC, &i) == -1) {
		PERROR("Could not set mixer source");
		close(fd);
		return -1;
	}
	close(fd);
	return 1;
}

int oss_core_setmixvol(char *cmd)
{
	int fd;
	char *cn;
	char *vol;
	int i, ivol;
	
	i = strcspn(cmd, ":");
	if(i == strlen(cmd)) {
		EPRINT("Invalid mixvol command\n");
		EPRINT("   eg: -mixvol /dev/mixer:line:50\n");
		return -1;
	}
	cmd[i] = 0;
	cn = cmd + i + 1;
	i = strcspn(cn, ":");
	if(i == strlen(cn)) {
		EPRINT("Invalid mixvol command\n");
		EPRINT("  eg: -mixvol /dev/mixer:line:50\n");
		return -1;
	}
	cn[i] = 0;
	vol = cn + i + 1;
	ivol = atoi(vol);
	if((ivol < 0) || (ivol > 100)) {
		EPRINT("Invlaid volume, value must be 0..100.\n");
		return -1;
	}
	ivol = (ivol << 8) | ivol;
	for(i = 0; i < MAX_DEV; i++) {
		if(strcmp(cn, oss_core_devnames[i]) == 0)
			break;
	}
	if(i == MAX_DEV) {
		EPRINT("Could not find mixer device \"%s\".\n", cn);
		return -1;
	}
	if ((fd = open(cmd, O_RDWR)) < 0) {
		PERROR("Could not open mixer device");
		return -1;
	}
	if(ioctl(fd, MIXER_WRITE(i), &ivol) == -1) {
		PERROR("Could not set mixer source");
		close(fd);
		return -1;
	}
	close(fd);
	return 1;
}

struct aud_core_t * oss_core_init(char ** args, ntime_t ftime)
{
	struct aud_core_t * act;
	struct oss_core_t * osst;
 	int i, j, t;
 	unsigned long long bpf;
 	pthread_t athread;
 	char * fdev = NULL;
 	int N = -1;
 	NV_ARGS;

	act = malloc(sizeof(struct aud_core_t));
	if(!act) {
		PERROR("Could not malloc new audio core");
		return NULL;
	}
	osst = malloc(sizeof(struct oss_core_t));
	if(!osst) {
		PERROR("Could not malloc new oss core");
		free(act);
		return NULL;
	}
	memcpy(act, &oss_core, sizeof(struct aud_core_t));
	act->priv = osst;
	memset(osst, 0, sizeof(struct oss_core_t));
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			osst->rate = atoi(NV_THIS_ARG);
			if(osst->rate < 8000) {
				EPRINT("Invalid argument to -r (>=8000).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			osst->bits = atoi(NV_THIS_ARG);
			if((osst->bits != 8) && (osst->bits != 16)) {
				EPRINT("Invalid argument to -b (8 or 16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			osst->chans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			osst->chans = 1;
		} else if(strcmp(NV_THIS_ARG, "-d") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -d.\n");
				goto usage;
			}
			fdev = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-mixsrc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -mixsrc.\n");
				goto usage;
			}
			if(oss_core_setmixsrc(NV_THIS_ARG) < 0) {
				EPRINT("Invalid argument to -mixsrc.\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-mixvol") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -mixvol.\n");
				goto usage;
			}
			if(oss_core_setmixvol(NV_THIS_ARG) < 0) {
				EPRINT("Invalid argument to -mixvol.\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-N") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -N.\n");
				goto usage;
			}
			N = atoi(NV_THIS_ARG);
			if(N < 4) {
				EPRINT("Invalid argument to -N (>=4).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			ftime = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
	}
	IPRINT("%d bps, %d bits, %s, from %s\n", osst->rate, osst->bits, (osst->chans == 1)?"mono":"stereo", fdev);

 	bpf = osst->rate * osst->chans * ((osst->bits==16)?2:1);
 	bpf *= ftime;
 	bpf /= 1000000000LL;
 	t = bpf / 4;
 	
	osst->fd = open(fdev, O_RDONLY);
	if(osst->fd<=0) {
		PERROR("Open oss device");
		goto error0;
	}
  
 	j = 1;
 	i = 2;
 	while((i*2) < t) {
 		i *= 2;
 		j++;
 	}
 	i = 65536 / i;
 	IPRINT("%lli bytes per frame (%d,%d)\n", bpf, i, j);
 	j = j | (i << 16);
	if (ioctl(osst->fd, SNDCTL_DSP_SETFRAGMENT, &j) < 0)
		PERROR("Could not set fragment size - will continue with default");

	j = osst->bits;
	if (ioctl(osst->fd, SNDCTL_DSP_SAMPLESIZE, &j) < 0) {
		PERROR("Could not set sample size");
		goto error1;
	}

	j = osst->chans - 1;
	if (ioctl(osst->fd, SNDCTL_DSP_STEREO, &j) < 0) {
		PERROR("WARNING could not set channels");
		goto error1;
	}
	if(j != (osst->chans - 1)) {
		EPRINT("WARNING could not set mono/stereo mode - please use -s or -m to override!\n");
		goto error1;
	}

	j = osst->rate;
	if (ioctl(osst->fd, SNDCTL_DSP_SPEED, &j) < 0) {
		PERROR("WARNING could not set speed, continuing with default");
	}
 	osst->bps = osst->rate * osst->chans * ((osst->bits==16)?2:1);
 	osst->realbps = j * osst->chans * ((osst->bits==16)?2:1);
	IPRINT("bps=%d, realbps=%d\n", osst->bps, osst->realbps);

	if (ioctl(osst->fd, SNDCTL_DSP_GETBLKSIZE, &j) < 0) {
		PERROR("Could not get block size");
		goto error1;
	}
	osst->fragsize = j;
	i = (N * bpf * 2) / j;
	IPRINT("%i buffer frags required of %i bytes\n", i, j);
	osst->frag = (struct frag_t *)malloc(sizeof(struct frag_t) * i);
	if(!osst->frag) {
		PERROR("Could not allocate fragment memory");
		goto error1;
	}
	memset(osst->frag, 0, sizeof(struct frag_t) * i);
	while(osst->nfrags < i) {
		(osst->frag + osst->nfrags)->buf = (unsigned char *)malloc(osst->fragsize);
		if(!(osst->frag + osst->nfrags)->buf) {
			if(osst->nfrags > 0) {
				PERROR("Could not allocate memory for all buffers, trying to continue anyway!");
			} else {
				PERROR("Could not allocate memory for any buffers - aborting!");
				goto error1;
			}
		}
		(osst->frag + osst->nfrags)->time = -1;
		(osst->frag + osst->nfrags)->full = 0;
		osst->nfrags++;
	}
	osst->head = 0;
	osst->tail = 0;
	osst->exit = 0;
	osst->done = 0;
	
	if(pthread_create(&athread, NULL, oss_core_thread, (void *)osst) != 0) {
		PERROR("Could not start ossio thread");
		goto error1;
	}
	return act;

error1:
	oss_core_freefrags(osst);
	close(osst->fd);
error0:
	free(act);
	free(osst);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(act);
	free(osst);
	return NULL;
}

struct aud_core_t oss_core = {
	aud_core_init: oss_core_init, 
	aud_core_close: oss_core_close,
	aud_core_tailtime: oss_core_tailtime,
	aud_core_droptail: oss_core_droptail,
	aud_core_read: oss_core_read,
	aud_core_bps: oss_core_bps,
	aud_core_overrides: NULL,
	aud_core_usage: oss_core_usage,
	priv: NULL
};
