/* 
   qtfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <string.h>
#include <lqt.h>
#include <colormodels.h>
#include "nvrec.h"
#ifdef HAVE_CONVERT
#include "convert.h"
#endif

struct qtfile_core_t {
	quicktime_t * fd;
	int abits;
	int achans;
	int width;
	int height;
	int fmt;
	unsigned char ** temp_rows;
#ifdef HAVE_CONVERT
	convert_t * conv;
#endif
};

struct file_core_t qtfile_core;

char *qtfile_core_args_fourcc(char **args)
{
	static char fourcc[5];
	char * compr = NULL;
 	NV_ARGS;
 	
	lqt_registry_init();
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG)
				break;
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				const lqt_codec_info_t *ci = NULL;
				int i, j;
				for(i = 0; i < lqt_get_num_video_codecs(); i++) {
					ci = lqt_get_video_codec_info(i);
					if(!ci) {
						EPRINT("Error retrieving codec info!\n");
						continue;
					}
					if(ci->type != LQT_CODEC_VIDEO)
						continue;
					if(ci->direction == LQT_DIRECTION_DECODE)
						continue;
					MPRINT("---------------------------------------------------------------------------\n");
					MPRINT("Name: %s\n", ci->name);
					MPRINT("Long name: %s\n", ci->long_name);
					MPRINT("Description: %s\n", ci->description);
					MPRINT("FourCC: ");
					for(j = 0; j < ci->num_fourccs; j++)
						MPRINT("%s ", ci->fourccs[j]);
					MPRINT("\n");
				}
				MPRINT("---------------------------------------------------------------------------\n");
				return NULL;
			}
			compr = NV_THIS_ARG;
		}
	}
	if(!compr) {
		snprintf(fourcc, 5, "RTJ0");
	} else {
		snprintf(fourcc, 5, "%s    ", compr);
	}
	return fourcc;
}

const lqt_codec_info_t *qtfile_core_args_codec(char **args)
{
	const lqt_codec_info_t *ci = NULL;
	int i, j;
	char * fourcc;
 	
 	fourcc = qtfile_core_args_fourcc(args);
 	if(!fourcc)
 		return NULL;
	DPRINT("Got fourcc: %s\n", fourcc);
	for(i = 0; i < lqt_get_num_video_codecs(); i++) {
		ci = lqt_get_video_codec_info(i);
		if(!ci) {
			EPRINT("Error retrieving codec info!\n");
			continue;
		}
		DPRINT("codec %d: %s\n", i, ci->name);
		for(j = 0; j < ci->num_fourccs; j++) {
			DPRINT("fourcc %d: %s\n", j, ci->fourccs[j]);
			if(strcmp(ci->fourccs[j], fourcc) == 0)
				break;
		}
		if(j != ci->num_fourccs)
			break;
		ci = NULL;
	}
	if(!ci) {
		EPRINT("Codec for fourcc \"%s\" not found!\n", fourcc);
		EPRINT("You may need to upgrade libquicktime to version >=0.9.1, or check your -vc argument.\n");
		return NULL;
	}
	return ci;
}

int qtfile_core_qtfmt_nvfmt(int qtfmt)
{
	switch(qtfmt) {
		case BC_BGR888:
		case BC_RGB888:
			return NV_RGB24;
		case BC_YUV422:
			return NV_YUV422;
		case BC_YUV420P:
			return NV_YUV420;
		default:
			return -1;
	}
	return -1;
}

void qtfile_core_close(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	DXPRINT("enter\n");
	if(!qtc)
		return;
	DXPRINT("quicktime_close\n");
	quicktime_close(qtc->fd);
	if(qtc->temp_rows)
		free(qtc->temp_rows);
#ifdef HAVE_CONVERT
	DXPRINT("free conv\n");
	if(qtc->conv) {
		convert_end(qtc->conv);
	}
#endif
	DXPRINT("free qtc\n");
	free(qtc);
	DXPRINT("free fct\n");
	free(fct);
	DXPRINT("leave\n");
}

void qtfile_core_nullvideo(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	
	if(!qtc)
		return;
	quicktime_write_frame(qtc->fd, NULL, 0, 0);
}

void qtfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	static int x, y, t;
	
	if(!qtc)
		return;
#ifdef HAVE_CONVERT
	if(qtc->conv) {
		buf = convert_convert(qtc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	if(!qtc->temp_rows) {
		switch(qtc->fmt) {
			case BC_YUV420P:
				qtc->temp_rows = malloc(sizeof(unsigned char *) * 3);
				break;
			case BC_BGR888:
			case BC_RGB888:
			case BC_YUV422:
				qtc->temp_rows = malloc(sizeof(unsigned char *) * qtc->height);
				break;
			default:
				EPRINT("Unsupported input video format (%d)!\n", qtc->fmt);
				break;
		}
	}
	switch(qtc->fmt) {
		case BC_YUV420P:
			qtc->temp_rows[0] = buf[0];
			qtc->temp_rows[1] = buf[1];
			qtc->temp_rows[2] = buf[2];
			break;
		case BC_RGB888:
			for(y = 0; y < qtc->height; y++) {
				for(x = 0; x < qtc->width; x++) {
					t = buf[y][x*3+0];
					buf[y][x*3+0] = buf[y][x*3+2];
					buf[y][x*3+2] = t;
				}
				qtc->temp_rows[y] = buf[y];
			}
			break;
		case BC_BGR888:
		case BC_YUV422:
			for(y = 0; y < qtc->height; y++) {
				qtc->temp_rows[y] = buf[y];
			}
			break;
		default:
			EPRINT("Unsupported input video format (%d)!\n", qtc->fmt);
			return;
	}
	quicktime_encode_video(qtc->fd, qtc->temp_rows, 0);
}

void qtfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	unsigned short *bs = (unsigned short *)buf;
	int i, s;

	if(!qtc)
		return;
	
	s = n >> ((qtc->abits == 16) ? 1 : 0);
	if(qtc->abits == 16) {
		for(i = 0; i < s; i++)
			bs[i] = ((bs[i]>>8)&0x00ff)|((bs[i]<<8)&0xff00);
	} else {
		for(i = 0; i < s; i++)
			buf[i] ^= 0x80;
	}

	s >>= (qtc->achans == 2) ? 1 : 0;
	quicktime_write_audio(qtc->fd, buf, s, 0);
}

void qtfile_core_sync(struct file_core_t * fct)
{
/*	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	fdatasync(fileno(qtc->fd->stream)); */
}

static unsigned int qtfile_core_default_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
#define MAX_FMTS 20
unsigned int *qtfile_core_format(char **args)
{
	static unsigned int formats[MAX_FMTS];
	const lqt_codec_info_t *ci;
	int i = 0, j, t;

	ci = qtfile_core_args_codec(args);
	if(!ci)
		return NULL;
	for(j = 0; j < ci->num_encoding_colormodels; j++) {
		t = qtfile_core_qtfmt_nvfmt(ci->encoding_colormodels[j]);
		if((t >= 0) && (i < (MAX_FMTS - 1))) {
			formats[i++] = t;
			DPRINT("Adding format %d\n", t);
		}
	}
#ifdef HAVE_CONVERT
	for(j = 0; i < (MAX_FMTS - 1); j++) {
		formats[i++] = qtfile_core_default_formats[j];
		if(formats[i++] == NV_END_FORMAT) {
			break;
		}
	}
#endif
 	formats[i] = NV_END_FORMAT;
	return formats;
}

void qtfile_core_usage()
{
	MPRINT("QUICKTIME file core usage:\n");
	MPRINT("\t-vc <fourcc | help for list>                   [RTJ0]\n");
	MPRINT("\t-vp <param name | help for list> <param value> [none]\n");
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"                [NONE]\n");
#endif
	MPRINT("\t-o <output filename>  [NONE]\n");
}

struct file_core_t * qtfile_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct qtfile_core_t * qtc;
	double f;
 	int arate = -1;
 	char *fname = NULL;
	const lqt_codec_info_t *ci = NULL;
	int i, j, k;
	int nvfmt = -1, qtfmt = -1;
	char * compr = NULL;
	char *vhook = NULL;
 	NV_ARGS;
 	
 	lqt_registry_init();
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new cap core");
 		return NULL;
 	}
 	qtc = malloc(sizeof(struct qtfile_core_t));
 	if(!qtc) {
 		PERROR("Could no malloc new qt core");
 		free(fct);
 		return NULL;
 	}
 	memset(qtc, 0, sizeof(struct qtfile_core_t));
 	memcpy(fct, &qtfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)qtc;
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			qtc->abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			qtc->achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			qtc->achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			qtc->width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			qtc->height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vc.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				MPRINT("Listing possible -vc options:\n");
				goto usage;
			} else {
				compr = NV_THIS_ARG;
			}
		} else if(strcmp(NV_THIS_ARG, "-vp") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vp.\n");
				goto usage;
			}
			if(!compr) {
				EPRINT("Use \"-vc <codec>\" before \"-vp\".\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				ci = qtfile_core_args_codec(args);
				if(!ci) {
					EPRINT("Could not find codec!\n");
					goto usage;
				}
				for(j = 0; j < ci->num_encoding_parameters; j++) {
					MPRINT("---------------------------------------------------------------------------\n");
					MPRINT("Name: %s\n", ci->encoding_parameters[j].name);
					MPRINT("Real Name: %s\n", ci->encoding_parameters[j].real_name);
					switch(ci->encoding_parameters[j].type) {
						case LQT_PARAMETER_INT:
							MPRINT("Integer parameter.\n");
							MPRINT("Default: %d\n", ci->encoding_parameters[j].val_default.val_int);
							MPRINT("Minimum: %d\n", ci->encoding_parameters[j].val_min);
							MPRINT("Maximum: %d\n", ci->encoding_parameters[j].val_max);
							break;
						case LQT_PARAMETER_STRING:
							MPRINT("String parameter.\n");
							MPRINT("Default: %s\n", ci->encoding_parameters[j].val_default.val_string);
							break;
						case LQT_PARAMETER_STRINGLIST:
							MPRINT("String parameter list.\n");
							MPRINT("Default: %s\n", ci->encoding_parameters[j].val_default.val_string);
							MPRINT("Options: ");
							for(k = 0; k < ci->encoding_parameters[j].num_stringlist_options; k++)
								MPRINT("\"%s\" ", ci->encoding_parameters[j].stringlist_options[k]);
							MPRINT("\n");
							break;
						default:
							MPRINT("Unhandled parameter type!\n");
							break;
					}
				}
				MPRINT("---------------------------------------------------------------------------\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
		}
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	IPRINT("File: %s\n", fname);
	
	ci = qtfile_core_args_codec(args);
	if(!ci) {
		EPRINT("Could not find codec!\n");
		goto error0;
	}
	for(i = 0; i < ci->num_encoding_colormodels; i++) {
		qtfmt = ci->encoding_colormodels[i];
		nvfmt = qtfile_core_qtfmt_nvfmt(qtfmt);
		if(nvfmt == format)
			break;
	}
	if(i == ci->num_encoding_colormodels) {
#ifdef HAVE_CONVERT
		IPRINT("No native format found - initialise format conversion\n");
		for(i = 0; i < ci->num_encoding_colormodels; i++) {
			qtfmt = ci->encoding_colormodels[i];
			nvfmt = qtfile_core_qtfmt_nvfmt(qtfmt);
			if(nvfmt >= 0) {
				break;
			}
		}
		if(nvfmt < 0) {
			EPRINT("Could not find any compatible format!\n");
			goto error0;
		}
		qtc->conv = convert_new(format, nvfmt, qtc->width, qtc->height, vhook);
		if(!qtc->conv) {
			EPRINT("Could not init format conversion core!\n");
		goto error0;
	}
#else
		EPRINT("Could not find a valid colour model and conversion not available!\n");
		goto error0;
#endif
	}
#ifdef HAVE_CONVERT
	if((!qtc->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
 		qtc->conv = convert_new(format, format, qtc->width, qtc->height, vhook);
 		if(!qtc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto error0;
 		}
	}
#endif
	qtc->fmt = qtfmt;
	IPRINT("Video: %s (%d, %d)\n", ci->name, qtfmt, nvfmt);
	
	unlink(fname);
	qtc->fd = quicktime_open(fname, 1, 1);
	if(!qtc->fd) {
		PERROR("Could not open file");
		goto error0;
	}
	f = 1e9 / (double)vrate;
	IPRINT("Frame rate is %f\n", f);
	quicktime_set_video(qtc->fd, 1, qtc->width, qtc->height, f, qtfile_core_args_fourcc(args));
	quicktime_set_audio(qtc->fd, qtc->achans, arate, qtc->abits, QUICKTIME_TWOS);
	quicktime_set_cmodel(qtc->fd, qtc->fmt);

	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-vp") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			for(j = 0; j < ci->num_encoding_parameters; j++) {
				if(strcmp(NV_THIS_ARG, ci->encoding_parameters[j].name) == 0) {
					if(!NV_NEXT_ARG) {
						EPRINT("Missing argument to -vp.\n");
						goto usage;
					}
					IPRINT("Setting: %s\n", ci->encoding_parameters[j].real_name);
					switch(ci->encoding_parameters[j].type) {
						case LQT_PARAMETER_INT:
							i = atoi(NV_THIS_ARG);
							if((i > ci->encoding_parameters[j].val_max) ||
							   (i < ci->encoding_parameters[j].val_min)) {
								EPRINT("Argument out of range!\n");
								goto usage;
							}
							IPRINT("New value: %d\n", i);
							quicktime_set_parameter(qtc->fd, ci->encoding_parameters[j].name, &i);
							break;
						case LQT_PARAMETER_STRING:
							IPRINT("New value: %s\n", NV_THIS_ARG);
							quicktime_set_parameter(qtc->fd, ci->encoding_parameters[j].name, NV_THIS_ARG);
							break;
						case LQT_PARAMETER_STRINGLIST:
							IPRINT("New value: %s\n", NV_THIS_ARG);
							quicktime_set_parameter(qtc->fd, ci->encoding_parameters[j].name, NV_THIS_ARG);
							break;
						default:
							MPRINT("Unhandled parameter type!\n");
							break;
					}
					j = ci->num_encoding_parameters + 1;
					break;
				}
			}
			if(j == ci->num_encoding_parameters) {
				EPRINT("Unrecognised parameter: %s\n", NV_THIS_ARG);
				goto error0;
			}
		}
	}

	return (void *)fct;

error0:
#ifdef HAVE_CONVERT
	if(qtc->conv) {
		convert_end(qtc->conv);
	}
#endif
	free(fct);
	free(qtc);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(qtc);
	return NULL;
}

struct file_core_t qtfile_core = {
	file_core_init: qtfile_core_init,
	file_core_close: qtfile_core_close,
	file_core_nullvideo: qtfile_core_nullvideo,
	file_core_fastvideo: qtfile_core_video,
	file_core_video: qtfile_core_video,
	file_core_audio: qtfile_core_audio,
	file_core_sync: qtfile_core_sync,
	file_core_format: qtfile_core_format,
	file_core_overrides: NULL,
	file_core_usage: qtfile_core_usage,
	priv: NULL
};
