/* 
   rtefile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <librte.h>
#include "nvrec.h"

struct rtefile_core_t {
	rte_context * context;

	rte_codec * acodec;
	double atime;
	double atimestep;
	rte_buffer abuf;

	rte_codec * vcodec;
	double vtime;
	double vtimestep;
	rte_buffer vbuf;

	unsigned char * abak;
	int abfill;
	int fragment_size;
};

void rtefile_core_usage()
{
	MPRINT("RTE file (mpeg-1) core usage:\n");
	MPRINT("\t-context <context or \"help\" for list>\n");
	MPRINT("\t-ac <audio codec or \"help\" for list>\n");
	MPRINT("\t-vc <audio codec or \"help\" for list>\n");
	MPRINT("\t-co <keyword> <value> - set context option\n");
	MPRINT("\t-aco <keyword> <value> - set audio codec option\n");
	MPRINT("\t-vco <keyword> <value> - set video codec option\n");
	MPRINT("\t-o <output filename>                  [NONE]\n");
}

struct file_core_t rtefile_core;

rte_context * rtefile_context(char ** args)
{
	rte_context_info * info;
	rte_context * context = NULL;
	const char * keyword = NULL;
	int i;
	NV_ARGS;
	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-context") == 0) {
			if(!NV_NEXT_ARG)
				return NULL;
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				MPRINT("Available rte contexts:\n");
				for(i = 0; (info = rte_context_info_enum(i)); i++) {
					MPRINT("--------------------------------------------------------------------------\n");
					MPRINT("  Keyword: %s\n", info->keyword);
					MPRINT("  Backend: %s\n", info->backend);
					MPRINT("  Label: %s\n", info->label);
					if(info->tooltip)
						MPRINT("  ToolTip: %s\n", info->tooltip);
					if(info->mime_type)
						MPRINT("  Mime Type: %s\n", info->mime_type);
					if(info->extension)
						MPRINT("  Extension: %s\n", info->extension);
					MPRINT("  Minimum Audio Streams: %d\n", info->min_elementary[RTE_STREAM_AUDIO]);
					MPRINT("  Maximum Audio Streams: %d\n", info->max_elementary[RTE_STREAM_AUDIO]);
					MPRINT("  Minimum Video Streams: %d\n", info->min_elementary[RTE_STREAM_VIDEO]);
					MPRINT("  Maximum Video Streams: %d\n", info->max_elementary[RTE_STREAM_VIDEO]);
					if(info->flags & RTE_FLAG_SEEKS)
						MPRINT("  Can Seek.\n");
				}
				return NULL;
			} else {
				keyword = NV_THIS_ARG;
			}
		}
	}
	if(!keyword) {
		info = rte_context_info_enum(0);
		if(info) {
			IPRINT("Defaulting to context %s.\n", info->label);
			keyword = info->keyword;
		}
	}
	if(!keyword) {
		EPRINT("Could not get any contexts!\n");
		return NULL;
	}
	context = rte_context_new(keyword, NULL, NULL);
	if(!context)
		EPRINT("Could not match context %s!\n", keyword);
	return context;
}

rte_codec * rtefile_codec(char ** args, rte_context * context, int vid)
{
	rte_codec_info * info;
	rte_codec * codec = NULL;
	const char * keyword = NULL;
	char * clo = (vid) ? "-vc" : "-ac";
	int i;
	NV_ARGS;
	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, clo) == 0) {
			if(!NV_NEXT_ARG)
				return NULL;
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				MPRINT("Available rte codecs:\n");
				for(i = 0; (info = rte_codec_info_enum(context, i)); i++) {
					if((info->stream_type != RTE_STREAM_VIDEO) && (vid))
						continue;
					if((info->stream_type != RTE_STREAM_AUDIO) && (!vid))
						continue;
					MPRINT("--------------------------------------------------------------------------\n");
					MPRINT("  Keyword: %s\n", info->keyword);
					MPRINT("  Label: %s\n", info->label);
					if(info->tooltip)
						MPRINT("  ToolTip: %s\n", info->tooltip);
				}
				return NULL;
			} else {
				keyword = NV_THIS_ARG;
			}
		}
	}
	if(!keyword) {
		for(i = 0; (info = rte_codec_info_enum(context, i)); i++) {
			if((info->stream_type != RTE_STREAM_VIDEO) && (vid))
				continue;
			if((info->stream_type != RTE_STREAM_AUDIO) && (!vid))
				continue;
			IPRINT("Defaulting to codec %s.\n", info->label);
			keyword = info->keyword;
			break;
		}
	}
	if(!keyword) {
		EPRINT("Could not get any %s codecs!\n", (vid)?"video":"audio");
		return NULL;
	}
	codec = rte_codec_set(context, keyword, 0, NULL);
	if(!codec)
		EPRINT("Could not match codec %s!\n", keyword);
	return codec;
}

rte_option_info * rtefile_genenum(rte_context * context, rte_codec * codec, int otype, int i)
{
	if(otype == 2)
		return rte_context_option_info_enum(context, i);
	return rte_codec_option_info_enum(codec, i);
}

int rtefile_options(char ** args, rte_context * context, rte_codec * codec, int otype)
{
	rte_option_info * info;
	const char * keyword = NULL;
	char * clo;
	char * type;
	int i, j;
	double id;
	char * is;
	rte_option_value val;
	NV_ARGS;
	
	switch(otype) {
		case 0: clo = "-aco"; type = "audio"; break;
		case 1: clo = "-vco"; type = "video"; break;
		case 2: clo = "-co"; type = "context"; break;
		default:
			EPRINT("Oops invalid option type!\n");
			return -1;
	}
	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, clo) == 0) {
			if(!NV_NEXT_ARG)
				return -1;
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				MPRINT("Available rte %s options:\n", type);
				for(i = 0; (info = rtefile_genenum(context, codec, otype, i)); i++) {
					MPRINT("--------------------------------------------------------------------------\n");
					MPRINT("  Keyword: %s\n", info->keyword);
					MPRINT("  Label: %s\n", info->label);
					if(info->tooltip)
						MPRINT("  ToolTip: %s\n", info->tooltip);
					if(info->menu.str) {
						MPRINT("  Menu option, default is %d.\n", info->def.num);
						MPRINT("  Possible values are:\n");
						switch(info->type) {
							case RTE_OPTION_INT:
								for(j = info->min.num; j <= info->max.num; j++)
									MPRINT("    %d: %d\n", j, info->menu.num[j]);
								break;
							case RTE_OPTION_REAL:
								for(j = info->min.num; j <= info->max.num; j++)
									MPRINT("    %d: %.4g\n", j, info->menu.dbl[j]);
								break;
							case RTE_OPTION_STRING:
							case RTE_OPTION_MENU:
								for(j = info->min.num; j <= info->max.num; j++)
									MPRINT("    %d: \"%s\"\n", j, info->menu.str[j]);
								break;
							default:
								MPRINT("  ** UNHANDLED OPTION TYPE **\n");
						}
						MPRINT("  (Use the number on the left when setting an option!)\n");
					} else {
						switch(info->type) {
							case RTE_OPTION_BOOL:
								MPRINT("  Boolean option (0 or 1), default is %d.\n", info->def.num);
								break;
							case RTE_OPTION_INT:
								MPRINT("  Integer option from %d to %d, default is %d.\n", info->min.num, info->max.num, info->def.num);
								break;
							case RTE_OPTION_REAL:
								MPRINT("  Float option from %.4g to %.4g, default is %.4g.\n", info->min.dbl, info->max.dbl, info->def.dbl);
								break;
							case RTE_OPTION_STRING:
								MPRINT("  String option, default is \"%s\".\n", info->def.str);
								break;
							default:
								MPRINT("  ** UNHANDLED OPTION TYPE **\n");
						}
					}
				}
				return -1;
			} else {
				keyword = NV_THIS_ARG;
				if(!NV_NEXT_ARG)
					return -1;
				break;
			}
		}
	}
	if(!keyword)
		return 0;
	if(otype == 2) {
		info = rte_context_option_info_keyword(context, keyword);
	} else {
		info = rte_codec_option_info_keyword(codec, keyword);
	}
	if(!info) {
		EPRINT("Could not find option %s!\n", keyword);
		return -1;
	}
	if(info->menu.str) {
		i = atoi(NV_THIS_ARG);
		DPRINT("Setting %s menu option %s to %d.\n", type, info->keyword, i);
		if(otype == 2) {
			rte_context_option_menu_set(context, keyword, i);
			rte_context_option_menu_get(context, keyword, &j);
		} else {
			rte_codec_option_menu_set(codec, keyword, i);
			rte_codec_option_menu_get(codec, keyword, &j);
		}
		if(i == j) {
			switch(info->type) {
				case RTE_OPTION_INT:
					IPRINT("Set %s menu option %s to %d.\n", type, info->keyword, info->menu.num[i]);
					break;
				case RTE_OPTION_REAL:
					IPRINT("Set %s menu option %s to %.4g.\n", type, info->keyword, info->menu.dbl[i]);
					break;
				case RTE_OPTION_STRING:
				case RTE_OPTION_MENU:
					IPRINT("Set %s menu option %s to \"%s\".\n", type, info->keyword, info->menu.str[i]);
					break;
				default:
					EPRINT("Unhandled option set type!\n");
					break;
			}
			return 0;
		} else {
			EPRINT("Could not set %s menu option %s to %d.\n", type, info->keyword, i);
			return -1;
		}
		return -1;
	}
	switch(info->type) {
		case RTE_OPTION_BOOL:
			i = atoi(NV_THIS_ARG);
			DPRINT("Setting %s bool option %s to %d.\n", type, info->keyword, i);
			if(otype == 2) {
				rte_context_option_set(context, keyword, i);
				rte_context_option_get(context, keyword, &val);
			} else {
				rte_codec_option_set(codec, keyword, i);
				rte_codec_option_get(codec, keyword, &val);
			}
			if(i == val.num) {
				IPRINT("Set %s bool option %s to %d.\n", type, info->keyword, i);
				return 0;
			} else {
				EPRINT("Could not set %s bool option %s to %d.\n", type, info->keyword, i);
				return -1;
			}
			break;
		case RTE_OPTION_INT:
			i = atoi(NV_THIS_ARG);
			DPRINT("Setting %s int option %s to %d.\n", type, info->keyword, i);
			if(otype == 2) {
				rte_context_option_set(context, keyword, i);
				rte_context_option_get(context, keyword, &val);
			} else {
				rte_codec_option_set(codec, keyword, i);
				rte_codec_option_get(codec, keyword, &val);
			}
			if(i == val.num) {
				IPRINT("Set %s int option %s to %d.\n", type, info->keyword, i);
				return 0;
			} else {
				EPRINT("Could not set %s int option %s to %d.\n", type, info->keyword, i);
				return -1;
			}
			break;
		case RTE_OPTION_REAL:
			id = atof(NV_THIS_ARG);
			DPRINT("Setting %s float option %s to %.4g.\n", type, info->keyword, id);
			if(otype == 2) {
				rte_context_option_set(context, keyword, id);
				rte_context_option_get(context, keyword, &val);
			} else {
				rte_codec_option_set(codec, keyword, id);
				rte_codec_option_get(codec, keyword, &val);
			}
			if(id == val.dbl) {
				IPRINT("Set %s float option %s to %.4g.\n", type, info->keyword, id);
				return 0;
			} else {
				EPRINT("Could not set %s float option %s to %.4g.\n", type, info->keyword, id);
				return -1;
			}
			break;
		case RTE_OPTION_STRING:
			is = NV_THIS_ARG;
			DPRINT("Setting %s string option %s to \"%s\".\n", type, info->keyword, is);
			if(otype == 2) {
				rte_context_option_set(context, keyword, is);
				rte_context_option_get(context, keyword, &val);
			} else {
				rte_codec_option_set(codec, keyword, is);
				rte_codec_option_get(codec, keyword, &val);
			}
			if(strcmp(is, val.str) == 0) {
				IPRINT("Set %s string option %s to \"%s\".\n", type, info->keyword, is);
				return 0;
			} else {
				EPRINT("Could not set %s float option %s to \"%s\".\n", type, info->keyword, is);
				return -1;
			}
			break;
		default:
			break;
	}
	EPRINT("Unhandled option set type!\n");
	return -1;
}

void rtefile_core_novideo(struct file_core_t * fct, unsigned char **buf)
{
}

void rtefile_core_nonullvideo(struct file_core_t * fct)
{
}

void rtefile_core_noaudio(struct file_core_t * fct, unsigned char *buf, int n)
{
}

struct file_core_t * rtefile_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct rtefile_core_t * rtec;
	int arate = -1;
	int abits = -1;
	int achans = -1;
	int width = -1;
	int height = -1;
	char *fname = NULL;
	int i;
	NV_ARGS;
	
	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
	if(!fct) {
		PERROR("Could no malloc new file core");
		return NULL;
	}
	memcpy(fct, &rtefile_core, sizeof(struct file_core_t));

	rtec = (struct rtefile_core_t *)malloc(sizeof(struct rtefile_core_t));
	if(!rtec) {
		PERROR("Could no malloc new rte core");
		free(fct);
		return NULL;
	}
	memset(rtec, 0, sizeof(struct rtefile_core_t));
	fct->priv = (void *)rtec;
	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	IPRINT("File: %s\n", fname);

	if(!(rtec->context = rtefile_context(args))) {
		EPRINT("Error initialising rte context.\n");
		goto error;
	}

	rtec->vcodec = rtefile_codec(args, rtec->context, 1);
	rtec->acodec = rtefile_codec(args, rtec->context, 0);
	if((!rtec->vcodec) && (!rtec->acodec)) {
		EPRINT("No audio or video codecs selected!\n");
		goto error;
	}
	if(rtefile_options(args, rtec->context, rtec->acodec, 2) != 0) {
		EPRINT("Error setting rtefile_options (audio).\n");
		goto error;
	}
 	if(rtec->vcodec) {
 		rte_stream_parameters param;
 		rte_pixfmt pixfmt;
 		
 		if(rtefile_options(args, rtec->context, rtec->vcodec, 1) != 0) {
 			EPRINT("Error setting rtefile_options (video).\n");
 			goto error;
 		}

 		switch(format) {
 			case NV_YUV420:	pixfmt = RTE_PIXFMT_YUV420; break;
 			case NV_YUV422: pixfmt = RTE_PIXFMT_YUYV; break;
 			default:
 				EPRINT("Unhandled pixel format!\n");
 				goto error;
	 	}
	 	param.video.pixfmt = pixfmt;
	 	param.video.frame_rate = 1e9 / (double)vrate;
	 	param.video.pixel_aspect = 1.0;
	 	param.video.width = width;
	 	param.video.height = height;
	 	param.video.offset = 0;
	 	param.video.u_offset = width * height;
	 	param.video.v_offset = width * height / 4;
	 	param.video.stride = width;
	 	param.video.uv_stride = width / 2;
	 	if(!rte_codec_parameters_set(rtec->vcodec, &param)) {
	 		EPRINT("Could not set video params!\n");
	 		goto error;
	 	}
	 	if((param.video.width != width) || (param.video.height != height)) {
	 		EPRINT("Could not match size, closest match is %d x %d.\n", param.video.width, param.video.height);
	 		goto error;
	 	}
	 	if(param.video.frame_rate != (1e9 / (double)vrate)) {
	 		EPRINT("Could not match frame rate, closest is %f.\n", param.video.frame_rate);
	 		goto error;
	 	}
	 	if(param.video.pixfmt != pixfmt) {
	 		EPRINT("Could not match pixel format - try a different codec!\n");
	 		goto error;
	 	}
		rtec->vtimestep = ((double)vrate) / 1e9;
		rtec->vtime = 0;
		i = 0;
		if(!rte_set_input_push_passive(rtec->vcodec, 10, &i)) {
			EPRINT("Could not set video push (%s)!\n", rte_errstr(rtec->context));
			goto error;
		}
		IPRINT("Got %d video buffers.\n", i);
 	} else {
 		fct->file_core_video = rtefile_core_novideo;
 		fct->file_core_nullvideo = rtefile_core_nonullvideo;
 		fct->file_core_fastvideo = rtefile_core_novideo;
 	}

 	if(rtec->acodec) {
 		rte_stream_parameters param;
 		rte_sndfmt sndfmt;
 		double tmp;
 		
 		if(rtefile_options(args, rtec->context, rtec->acodec, 0) != 0) {
 			EPRINT("Error setting rtefile_options (audio2).\n");
 			goto error;
 		}

 		switch(abits) {
 			case 16: sndfmt = RTE_SNDFMT_S16_LE; break;
 			case 8: sndfmt = RTE_SNDFMT_U8; break;
 			default:
 				EPRINT("Unhandled sound format!\n");
 				goto error;
	 	}
	 	param.audio.sndfmt = sndfmt;
	 	param.audio.sampling_freq = arate;
	 	param.audio.channels = achans;
	 	param.audio.fragment_size = 0;
	 	if(!rte_codec_parameters_set(rtec->acodec, &param)) {
	 		EPRINT("Could not set audio params!\n");
	 		goto error;
	 	}
	 	if(param.audio.sndfmt != sndfmt) {
	 		EPRINT("Could not match format, try a different bit depth!\n");
	 		goto error;
	 	}
	 	if(param.audio.sampling_freq != arate) {
	 		EPRINT("Could not match sample rate, closest is %d.\n", param.audio.sampling_freq);
	 		goto error;
	 	}
	 	if(param.audio.channels != achans) {
	 		EPRINT("Could not match mono/stereo setting, required channels is %d!\n", param.audio.channels);
	 		goto error;
	 	}
	 	IPRINT("Got audio block size %d\n", param.audio.fragment_size);
	 	rtec->fragment_size = param.audio.fragment_size;
	 	tmp = arate;
	 	tmp *= achans;
	 	tmp *= (abits==16)?2:1; /* bytes/sec */
		rtec->atimestep = (double)rtec->fragment_size / tmp;
		rtec->atime = 0;
		rtec->abak = malloc(rtec->fragment_size * 2);
		if(!rtec->abak) {
			EPRINT("Could not malloc abak!\n");
			goto error;
		}
		i = 0;
		if(!rte_set_input_push_passive(rtec->acodec, 10, &i)) {
			EPRINT("Could not set audio push!\n");
			goto error;
		}
		IPRINT("Got %d audioo buffers.\n", i);
 	} else {
 		fct->file_core_audio = rtefile_core_noaudio;
 	}
 	
 	if(strcmp(fname, "-") == 0) {
 		IPRINT("Setting output to stdout!\n");
 		if(!rte_set_output_stdio(rtec->context, STDOUT_FILENO)) {
 			EPRINT("Could not set stdout for output!\n");
 			goto error;
 		}
 	} else {
 		IPRINT("Setting output to file \"%s\"!\n", fname);
 		if(!rte_set_output_file(rtec->context, fname)) {
 			EPRINT("Could not set file for output!\n");
 			goto error;
 		}
 	}
 	
 	if(!rte_start(rtec->context, 0.0, NULL, TRUE)) {
 		EPRINT("Could not start encoding (%s)!\n", rte_errstr(rtec->context));
 		goto error;
 	}

	return fct;

error:
	if(rtec->acodec)
		rte_codec_remove_codec(rtec->acodec);
	if(rtec->vcodec)
		rte_codec_remove_codec(rtec->vcodec);
	if(rtec->context)
		rte_context_delete(rtec->context);
	free(fct);
	free(rtec);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(rtec);
	return NULL;
}

void rtefile_core_close(struct file_core_t * fct)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	
	if(!rtec)
		return;
	rte_stop(rtec->context, rtec->vtime);

	if(rtec->acodec)
		rte_codec_remove_codec(rtec->acodec);
	if(rtec->vcodec)
		rte_codec_remove_codec(rtec->vcodec);
	if(rtec->context)
		rte_context_delete(rtec->context);

	free(rtec);
	free(fct);
}

void rtefile_core_nullvideo(struct file_core_t * fct)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	if(!rtec)
		return;
	if(!rte_push_buffer(rtec->vcodec, NULL, FALSE))
		EPRINT("Error sending NULL frame!\n");
	rtec->vtime += rtec->vtimestep;
}

void rtefile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	
	if(!rtec)
		return;
	if(!rtec->vbuf.data) {
		rte_push_buffer(rtec->vcodec, &rtec->vbuf, FALSE);
		if(!rtec->vbuf.data) {
			EPRINT("Could not get new buffer!\n");
			return;
		}
	}
	DPRINT("%d, %d, %f\n", (unsigned int)rtec->vbuf.data, rtec->vbuf.size, rtec->vtime);
	memcpy(rtec->vbuf.data, buf[0], rtec->vbuf.size);
	rtec->vbuf.timestamp = rtec->vtime;
	if(!rte_push_buffer(rtec->vcodec, &rtec->vbuf, FALSE))
		EPRINT("Error sending video buffer!\n");
	rtec->vtime += rtec->vtimestep;
}

void rtefile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	int c;

	if(!rtec)
		return;

	if(!rtec->abuf.data) {
		rte_push_buffer(rtec->acodec, &rtec->abuf, FALSE);
		if(!rtec->abuf.data) {
			EPRINT("Could not get new buffer!\n");
			return;
		}
	}

	DPRINT("%d, %d, %f\n", (unsigned int)rtec->abuf.data, rtec->abuf.size, rtec->atime);

	while((rtec->abfill + n) >= rtec->abuf.size) {
		if(rtec->abfill > 0)
			memcpy(rtec->abuf.data, rtec->abak, rtec->abfill);
		c = rtec->abuf.size - rtec->abfill;
		memcpy(rtec->abuf.data + rtec->abfill, buf, c);
		rtec->abfill = 0;
		n -= c;
		buf += c;
		if(!rte_push_buffer(rtec->acodec, &rtec->abuf, FALSE))
			EPRINT("Error sending audio buffer!\n");
		rtec->atime += rtec->atimestep;
	}

	if(n) {
		memcpy(rtec->abak + rtec->abfill, buf, n);
		rtec->abfill += n;
	}
}

void rtefile_core_sync(struct file_core_t * fct)
{
	return;
}

static unsigned int rtefile_core_formats[] = {NV_YUV420, NV_YUV422, NV_END_FORMAT};

unsigned int *rtefile_core_format(char **args)
{
	return rtefile_core_formats;
}

void rtefile_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t rtefile_core = {
	file_core_init: rtefile_core_init,
	file_core_close: rtefile_core_close,
	file_core_nullvideo: rtefile_core_nullvideo,
	file_core_fastvideo: rtefile_core_video,
	file_core_video: rtefile_core_video,
	file_core_audio: rtefile_core_audio,
	file_core_sync: rtefile_core_sync,
	file_core_format: rtefile_core_format,
	file_core_overrides: rtefile_core_overrides,
	file_core_usage: rtefile_core_usage,
	priv: NULL
};
