/* 
   v4l1_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h> 
#include <fcntl.h> 
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h> 
#include <errno.h>    
#include <unistd.h>   
#include <linux/videodev.h>
#include <pthread.h>
#include <signal.h>

#include "nvrec.h"

struct vbuf_t {
	unsigned char **buf;	// pointer to memory of each buffer
	char full;		// buffer SYNC'ed
	ntime_t ts;		// buffer SYNC'ed
	unsigned char *mem;	// memory if we need to bounce it
};

struct v4l1_core_t {
	int fd;			// file descriptor
	
	int width;		// image width we got
	int height;		// image height we got
	ntime_t frame_period;	// time for each frame
	int planes;		// number of planes/rows per image
	int framesize;		// size of frames
	
 	struct video_mbuf vmb;	// mmap buffers
	int nbufs;		// number of buffers received
	unsigned char *mmbuf;	// memory map buffer
	struct vbuf_t *vbufs;	// the individual buffers

	int last_index;		// index of last frame dequeued
	int last_seq;		// last sequence number
	ntime_t last_time;	// timestamp of last frame
	int new_index;		// index of frame dequeued
	int new_seq;		// new sequence number
	ntime_t new_time;	// timestamp of new frame
	
	ntime_t start_time;	// timestamp of first frame
	
	struct video_mmap mm;	// temp buffer
	
	int exit;		// stop SYNC thread
	int done;		// SYNC thread terminated

	int sync;		// sync end
	int queue;		// queue end
	int read;		// read point
	
	int copyread;		// will copy buffers
};

void * v4l1copy_core_thread(void * t)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)t;
	int i;
	int frame = 0;
	ntime_t ttime;
	struct sched_param sp;
	sigset_t newmask;
	
	sigfillset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, NULL) != 0) {
		PERROR("Could not block signals, but continuing anyway");
	} 
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				PERROR("Could not get any realtime priority");
				EPRINT("\t- This is not a serious problem, but capture quality under high\n");
				EPRINT("\t  load could be improved by running as root.\n");
			}
		}
	}
	
	while(!vct->exit) {
		i = -1;
		while(i < 0) {
			i = ioctl(vct->fd, VIDIOCSYNC, &frame);
			if(i < 0 && errno == EINTR) continue;
			if(i < 0) {
				PERROR("Error dequeueing buffer (trying to continue)");
				vct->mm.frame = frame;
				ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm);
				frame++;
				if(frame >= vct->vmb.frames)
					frame = 0;
			}
		}
		ttime = getntime();
		if((vct->vbufs + vct->sync)->full == 1) { // Nothing left to sync!
			EPRINT("No place to put new data - discarding %d\n", vct->exit);
			vct->mm.frame = frame;
			if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
				PERROR("Could not queue new buffer");
			frame++;
			if(frame >= vct->vmb.frames)
				frame = 0;
			usleep(vct->frame_period / 1000);
			continue;
		}
		memcpy((vct->vbufs + vct->sync)->mem, vct->mmbuf + vct->vmb.offsets[frame], vct->framesize);
		vct->mm.frame = frame;
		if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
			PERROR("Could not queue new buffer");
		frame++;
		if(frame >= vct->vmb.frames)
			frame = 0;
		(vct->vbufs + vct->sync)->ts = ttime;
		(vct->vbufs + vct->sync)->full = 1;
		vct->sync++;
		if(vct->sync >= vct->nbufs)
			vct->sync = 0;
	}
	DXPRINT("main loop done\n");
	IPRINT("Exiting!\n");
	DXPRINT("signal loop done\n");
	vct->done = 1;
	DXPRINT("leave\n");
	return NULL;
}

void * v4l1_core_thread(void * t)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)t;
	int i;
	struct sched_param sp;
	sigset_t newmask;
	
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGINT);sigemptyset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, NULL) != 0) {
		PERROR("Could not block signals, but continuing anyway");
	}
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		PERROR("Could not get FIFO priority");
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			PERROR("Could not get RR priority");
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				PERROR("Could not get any priority");
			}
		}
	}
	while(!vct->exit) {
		if((vct->vbufs + vct->sync)->full == 1) { // Nothing left to sync!
			 EPRINT("All buffers SYNCed - discarding data %d\n", vct->exit);
			 usleep(vct->frame_period / 1000);
			 continue;
		}
		i = -1;
		while(i < 0) {
			i = ioctl(vct->fd, VIDIOCSYNC, &vct->sync);
			if(i < 0 && errno == EINTR) continue;
			if(i < 0) {
				PERROR("Error dequeueing buffer (trying to continue)");
				i = 0;
			}
		}
		(vct->vbufs + vct->sync)->ts = getntime();
		(vct->vbufs + vct->sync)->full = 1;
		vct->sync++;
		if(vct->sync >= vct->nbufs)
			vct->sync = 0;
	}
	DXPRINT("main loop done\n");
	IPRINT("Exiting!\n");
	DXPRINT("signal loop done\n");
	vct->done = 1;
	DXPRINT("leave\n");
	return NULL;
}

void v4l1_core_freevbufs(volatile struct v4l1_core_t * vct)
{
	DXPRINT("enter\n");
	if(!vct->vbufs)
		return;
	DXPRINT("free buf loop\n");
	while(vct->nbufs > 0) {
		vct->nbufs--;
		if((vct->vbufs + vct->nbufs)->buf)
			free((vct->vbufs + vct->nbufs)->buf);
		if((vct->vbufs + vct->nbufs)->mem)
			free((vct->vbufs + vct->nbufs)->mem);
	}
	DXPRINT("free vbufs\n");
	free(vct->vbufs);
	DXPRINT("leave\n");
}

void v4l1_core_usage()
{
	MPRINT("v4l1 capture core usage:\n");
	MPRINT("\t-v <device name> capture device name  [/dev/video]\n");
	MPRINT("\t-w <width> picture width              [384]\n");
	MPRINT("\t-h <height> picture height            [288]\n");
	MPRINT("\t-norm <PAL|NTSC|SECAM> set input norm [unchanged]\n");
	MPRINT("\t-input <input name> set input         [unchanged]\n");
	MPRINT("\t-frequency <MHz> set input frequency  [unchanged]\n");
}

struct cap_core_t * v4l1_core_init(char ** args, unsigned int **pfmt_p)
{
	struct cap_core_t * cct;
	struct v4l1_core_t * vct;
 	int i, j, n;
 	pthread_t vthread;
 	struct video_window vwin;
 	struct video_audio vaud;
	char * fdev = NULL;
	int N = -1;
	char * norm = NULL;
	char * inp = NULL;
	double freq = -1;
	unsigned int *pfmt = *pfmt_p;
	NV_ARGS;

	cct = malloc(sizeof(struct cap_core_t));
	if(!cct) {
		PERROR("Could no malloc new cap core");
		return NULL;
	}
	vct = malloc(sizeof(struct v4l1_core_t));
	if(!vct) {
		PERROR("Could no malloc new v4l2 core");
		free(cct);
		return NULL;
	}
	memcpy(cct, &v4l1_core, sizeof(struct cap_core_t));
	cct->priv = (void *)vct;
	memset(vct, 0, sizeof(struct v4l1_core_t));
 
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			vct->width = atoi(NV_THIS_ARG);
			if(vct->width < 16) {
				EPRINT("Invalid argument to -w (>=16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			vct->height = atoi(NV_THIS_ARG);
			if(vct->height < 16) {
				EPRINT("Invalid argument to -h (>=16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-v") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -v.\n");
				goto usage;
			}
			fdev = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-N") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -N.\n");
				goto usage;
			}
			N = atoi(NV_THIS_ARG);
			if(N < 4) {
				EPRINT("Invalid argument to -N (>=4).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-norm") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -norm.\n");
				goto usage;
			}
			norm = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-input") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -input.\n");
				goto usage;
			}
			inp = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-frequency") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -frequency.\n");
				goto usage;
			}
			freq = atof(NV_THIS_ARG);
			if(freq < 0) {
				EPRINT("Invalid argument to -frequency (>=0).\n");
				goto usage;
			}
		}
	}
	IPRINT("Size: %dx%d from %s\n", vct->width, vct->height, fdev);

	vct->fd = open(fdev, O_RDWR);
	if(vct->fd<=0) {
		PERROR("Open capture device");
		goto error0;
	}

	if(inp) {
		struct video_channel chan;
		int i = 0;
		
		do {
			chan.channel = i;
			if(ioctl(vct->fd, VIDIOCGCHAN, &chan)) {
				EPRINT("Unknown input '%s'\n", inp);
				goto error1;
			}
			if(strcmp(chan.name, inp) == 0)
				break;
			i++;
		}  while (i > 0);
		if(norm) {
			if(strcmp(norm, "PAL") == 0) {
				chan.norm = VIDEO_MODE_PAL;
			} else if(strcmp(norm, "NTSC") == 0) { 
				chan.norm = VIDEO_MODE_NTSC;
			} else if(strcmp(norm, "SECAM") == 0) { 
				chan.norm = VIDEO_MODE_SECAM;
			} else {
				EPRINT("Unknown norm '%s'\n", norm);
				goto error1; 
			}
		}
		if(ioctl(vct->fd, VIDIOCSCHAN, &chan)) {
			PERROR("Could not set input");
			goto error1;
		}
		switch(chan.norm) {
			case VIDEO_MODE_SECAM:
			case VIDEO_MODE_PAL:
				vct->frame_period = 40000000ULL;
				break;
			case VIDEO_MODE_NTSC:
				vct->frame_period = 33366700ULL;
				break;
		}
	} else if(norm) {
		EPRINT("Sorry, v4l1 sucks.  If you want to set the norm, you must set the input too.\n");
		goto error1;
	} else {
		vct->frame_period = 40000000ULL; /* Default to PAL */
	}
	
	if(freq > 0) {
		int i = (int)(freq * 16.0);
		if(ioctl(vct->fd, VIDIOCSFREQ, &i)) {
			PERROR("Could not set frequency");
			goto error1;
		}
	}
 
 	vwin.x = 0;
 	vwin.y = 0;
 	vwin.flags = 0;
 	vwin.width = vct->width;
 	vwin.height = vct->height;
 	vwin.clips = NULL;
 	vwin.clipcount = 0;
	if (ioctl(vct->fd, VIDIOCSWIN, &vwin)) {
		PERROR("Set window");
		goto error1;
	}
	vct->mm.width = vct->width;
	vct->mm.height = vct->height;
	if (ioctl(vct->fd, VIDIOCGMBUF, &vct->vmb) != 0) {
		PERROR("Could not request buffers");
		goto error1;
	}
	if (vct->vmb.frames == 0) {
		EPRINT("Could not get any buffers\n");
		goto error1;
	}
	if (vct->vmb.frames < N) {
		EPRINT("Could not get enough buffers - using copy mode\n");
		n = N;
		vct->copyread = 1;
	} else {
		n = vct->vmb.frames;
	}
	vct->mmbuf = (unsigned char *)mmap(NULL, vct->vmb.size, PROT_READ|PROT_WRITE, MAP_SHARED, vct->fd, 0);
	if((int)(vct->mmbuf) == -1) {
		PERROR("MMAP failed");
		goto error1;
	}

	IPRINT("Got %i buffers\n", vct->vmb.frames);
	if(vct->copyread)
		IPRINT("Emulating %i buffers\n", n);

	vct->vbufs = (struct vbuf_t *)malloc(sizeof(struct vbuf_t) * n);
	if(!vct->vbufs) {
		EPRINT("Could not allocate memory (vbufs)");
		goto error1;
	}
	memset(vct->vbufs, 0, sizeof(struct vbuf_t) * n);

	while (*pfmt != NV_END_FORMAT) {
		switch (*pfmt) {
			case NV_RGB24:
				IPRINT("Trying video format RGB24\n");
				vct->mm.format = VIDEO_PALETTE_RGB24;
				vct->planes = vct->height;
				vct->framesize = vct->width * vct->height * 3;
				break;
			case NV_YUV420:
				IPRINT("Trying video format YUV420\n");
				vct->mm.format = VIDEO_PALETTE_YUV420P;
				vct->planes = 3;
				vct->framesize = (vct->width * vct->height * 3) / 2;
				break;
			case NV_YUV422:
				IPRINT("Trying video format YUV422\n");
				vct->mm.format = VIDEO_PALETTE_YUV422;
				vct->planes = vct->height;
				vct->framesize = vct->width * vct->height * 2;
				break;
			default:
				IPRINT("Unsupported video format\n");
				pfmt++;
				continue; 
		}
		for(i = 0; i < vct->vmb.frames; i++) {
			vct->mm.frame = i;
			if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0) {
				PERROR("Could not queue new buffer");
				break;
			}
		}
		if(i < vct->vmb.frames) {
			pfmt++;
			continue;
		}
		break;
	} 
	if(*pfmt == NV_END_FORMAT) {
		EPRINT("Could not find valid format\n");
		goto error1;
	}
	*pfmt_p = pfmt;

	while(vct->nbufs < n) {
		(vct->vbufs + vct->nbufs)->buf = (unsigned char **)malloc(sizeof(unsigned char *) * vct->planes);
		if(!(vct->vbufs + vct->nbufs)->buf) {
			PERROR("Could not allocate buffer pointers");
			goto error1;
		}
		if(vct->copyread) {
			(vct->vbufs + vct->nbufs)->mem = (unsigned char *)malloc(vct->framesize);
			if(!(vct->vbufs + vct->nbufs)->mem) {
				PERROR("Could not allocate buffer memory");
				goto error1;
			}
			(vct->vbufs + vct->nbufs)->buf[0] = (vct->vbufs + vct->nbufs)->mem;
		} else {
			(vct->vbufs + vct->nbufs)->buf[0] = vct->mmbuf + vct->vmb.offsets[vct->nbufs];
		}
		switch(*pfmt) {
			case NV_YUV420:
				(vct->vbufs + vct->nbufs)->buf[1] =
					(vct->vbufs + vct->nbufs)->buf[0] +
					vct->width * vct->height;
				(vct->vbufs + vct->nbufs)->buf[2] =
					(vct->vbufs + vct->nbufs)->buf[1] +
					(vct->width * vct->height) / 4;
				break;
			case NV_RGB24:
				for(j = 1; j < vct->height; j++) {
					(vct->vbufs + vct->nbufs)->buf[j] =
						(vct->vbufs + vct->nbufs)->buf[j - 1]
						+ vct->width * 3;
				}
				break;
			case NV_YUV422:
				for(j = 1; j < vct->height; j++) {
					(vct->vbufs + vct->nbufs)->buf[j] =
						(vct->vbufs + vct->nbufs)->buf[j - 1]
						+ vct->width * 2;
				}
				break;
		}
		vct->nbufs++;
	}
	vct->last_index = -1;
	vct->last_seq = -1;
	vct->new_index = -1;
	vct->new_seq = -1;
	
	vct->exit = 0;
	vct->done = 0;
	vct->sync = 0;
	vct->queue = 0; 
	vct->read = 0;
	
	memset(&vaud, 0, sizeof(vaud));
	if(ioctl(vct->fd, VIDIOCGAUDIO, &vaud)!=0)
		PERROR("Could not get mute status");
	vaud.flags &= ~VIDEO_AUDIO_MUTE;
	if(ioctl(vct->fd, VIDIOCSAUDIO, &vaud)!=0)
		PERROR("Could not set mute status");

	if(vct->copyread) {
		if(pthread_create(&vthread, NULL, v4l1copy_core_thread, (void *)vct) != 0) {
			PERROR("Could not start v4l1 thread");
			goto error1;
		}
	} else {
		if(pthread_create(&vthread, NULL, v4l1_core_thread, (void *)vct) != 0) {
			PERROR("Could not start v4l1 thread");
			goto error1;
		}	
	}
	return cct;
error1:
	v4l1_core_freevbufs(vct);
	close(vct->fd);
error0:
	free(cct);
	free(vct);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(cct);
	free(vct);
	return NULL;
}

void v4l1_core_close(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;
 	struct video_audio vaud;

	DXPRINT("enter\n");
	if(!vct)
		return;
	DXPRINT("signal core exit\n");
	vct->exit = 1;
	DXPRINT("wait for core exit\n");
	while(!vct->done)
		usleep(100000);
	DXPRINT("core exited\n");
	IPRINT("core thread terminated.\n");

	DXPRINT("blank vaud\n");
	memset(&vaud, 0, sizeof(vaud));
	DXPRINT("set vaud\n");
	if(ioctl(vct->fd, VIDIOCGAUDIO, &vaud)!=0)
		PERROR("Could not get mute status");
	DXPRINT("setup mute\n");
	vaud.flags |= VIDEO_AUDIO_MUTE;
	DXPRINT("do mute\n");
	if(ioctl(vct->fd, VIDIOCSAUDIO, &vaud)!=0)
		PERROR("Could not set mute status");

	DXPRINT("free bufs\n");
	v4l1_core_freevbufs(vct);
	DXPRINT("close fd\n");
	close(vct->fd);
	DXPRINT("free vct\n");
	free((void *)vct);
	DXPRINT("free cct\n");
	free(cct);
	DXPRINT("leave\n");
	return;
}

unsigned char ** v4l1_core_next(struct cap_core_t * cct, ntime_t * ftime)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;

	if (!vct)
		return NULL;
	if (vct->new_index < 0) { // FIXME get out of this loop on INT
		while((vct->vbufs + vct->read)->full == 0) {
			usleep(10000); // wait for new frame
		}
		vct->new_index = vct->read;
		vct->new_time = (vct->vbufs + vct->new_index)->ts;
		if(vct->last_index < 0) {
			vct->start_time = vct->new_time;
			vct->new_seq = 0;
		} else {
			vct->new_seq =
				( (vct->new_time - vct->start_time)
				+ (vct->frame_period >> 1) ) / 
				vct->frame_period;
		}
		vct->read++;
		if(vct->read >= vct->nbufs)
			vct->read = 0;
	}
	if(vct->last_index < 0) {
		/* First frame */
		if (vct->new_index < 0) return NULL;
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		vct->new_index = -1;
		*ftime = vct->last_time;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	} else if(vct->new_seq == (vct->last_seq + 1)) {
		/* New frame, in sequence */
		if(!vct->copyread) {
			vct->mm.frame = vct->last_index;
			if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
				PERROR("Could not requeue old buffer");
		}
		(vct->vbufs + vct->last_index)->full = 0;
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		*ftime = vct->last_time;
		vct->new_index = -1;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	} else if(vct->new_seq <= vct->last_seq)
	{
		/* New frame out of sequence! */
		/* FIXME - we can get some strange frame ordering here! */
		EPRINT("New frame out of sequence - ignoring\n");
		if(vct->new_index >= 0) {
			if(!vct->copyread) {
				vct->mm.frame = vct->last_index;
				if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
					PERROR("Could not requeue old buffer");
			}
			(vct->vbufs + vct->last_index)->full = 0;
			vct->last_index = vct->new_index;
			vct->new_index = -1;
		}
		return NULL;
	} else {
		/* New frame - lost something between though */
		vct->last_seq++;
		vct->last_time += vct->frame_period;
		*ftime = vct->last_time;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	}
	return NULL;
}

double v4l1_core_status(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;
	double qd;

	if(!vct)
		return -1;
	if(vct->new_index >= 0) {
		qd = vct->new_time;
	} else if(vct->last_index >= 0) {
		qd = vct->last_time;
	} else {
		return 0;
	}
	qd = (double)getntime() - qd;
	qd /= (double)vct->nbufs * (double)vct->frame_period;
	return qd * 100.0;
}

inline ntime_t v4l1_core_ftime(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;

	if(!vct)
		return -1;
	return vct->frame_period;
}

struct cap_core_t v4l1_core = {
	cap_core_init: v4l1_core_init,
	cap_core_close: v4l1_core_close,
	cap_core_next: v4l1_core_next,
	cap_core_status: v4l1_core_status,
	cap_core_ftime: v4l1_core_ftime,
	cap_core_overrides: NULL,
	cap_core_usage: v4l1_core_usage,
	priv: NULL
};
