/* 
   v4l2_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <linux/videodev.h>

#include "nvrec.h"

#ifdef V4L2_MAJOR_VERSION

struct v4l2_core_t {
	int fd;			// file descriptor
	
	int width;		// image width we got
	int height;		// image height we got
	unsigned int pfmt;	// pixel format we got
	ntime_t frame_period;	// time for each frame
	int planes;		// number of planes/rows per image
	
	int nbufs;		// number of buffers received
	unsigned char **bufs;	// pointer to memory of each buffer

	int last_index;		// index of last frame dequeued
	int last_seq;		// last sequence number
	ntime_t last_time;	// timestamp of last frame
	int new_index;		// index of frame dequeued
	int new_seq;		// new sequence number
	ntime_t new_time;	// timestamp of new frame
	
	ntime_t start_time;	// timestamp of first frame
	
	int old_mute;		// old value of mute setting
	
	struct v4l2_buffer cbuf; // temp buffer
};

// The maximum number of buffers to allocate		 

void v4l2_core_close(struct cap_core_t * cct)
{
	struct v4l2_core_t * vct = (struct v4l2_core_t *)cct->priv;
	struct v4l2_control ctrl;
	int i;

	DXPRINT("enter\n");
	if(!vct)
		return;
	DXPRINT("setup mute\n");
	memset(&ctrl, 0, sizeof(ctrl));
	ctrl.id = V4L2_CID_AUDIO_MUTE;  
	ctrl.value = vct->old_mute;
	DXPRINT("do mute\n");
	ioctl(vct->fd, VIDIOC_S_CTRL, &ctrl);

	DXPRINT("setup streamoff\n");
	i = V4L2_BUF_TYPE_CAPTURE;
	DXPRINT("do streamoff\n");
	ioctl(vct->fd, VIDIOC_STREAMOFF, &i);
	
	if(vct->bufs)
		free(vct->bufs);

	DXPRINT("close fd\n");
	close(vct->fd);
	DXPRINT("free vct\n");
	free(vct);
	DXPRINT("free cct\n");
	free(cct);
	DXPRINT("leave\n");
	return;
}

unsigned char ** v4l2_core_next(struct cap_core_t * cct, ntime_t * ftime)
{
	struct v4l2_core_t * vct = (struct v4l2_core_t *)cct->priv;

	if(!vct)
		return NULL;
	if (vct->new_index < 0) {
		int r = -1;
		fd_set fds;
		while(r <= 0) {
			FD_ZERO(&fds);
			FD_SET(vct->fd, &fds);
			r = select(vct->fd + 1, &fds, NULL, NULL, NULL);
			if (r < 0 && errno == EINTR) continue;
			if(r<0)
				PERROR("Select error");
		}	
		vct->cbuf.type = V4L2_BUF_TYPE_CAPTURE;
		if(ioctl(vct->fd, VIDIOC_DQBUF, &vct->cbuf)!=0) {
			PERROR("Could not dequeue new buffer");
			vct->new_index = -1;
			vct->new_seq = -1;
		} else {
			vct->new_index = vct->cbuf.index;
			if(vct->last_index < 0) {
				vct->start_time = vct->cbuf.timestamp;
				vct->new_seq = 0;
			} else {
				vct->new_seq =
					( (vct->cbuf.timestamp - vct->start_time)
					+ (vct->frame_period >> 1) ) / 
					vct->frame_period;
			}
			vct->new_time = vct->cbuf.timestamp;
		}
	}
	if(vct->last_index < 0) {
		/* First frame */
		if (vct->new_index < 0) return NULL;
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		vct->new_index = -1;
		*ftime = vct->last_time;
		return &vct->bufs[vct->last_index * vct->planes];
	} else if(vct->new_seq == (vct->last_seq + 1)) {
		/* New frame, in sequence */
		vct->cbuf.index = vct->last_index;
		if(ioctl(vct->fd, VIDIOC_QBUF, &vct->cbuf)!=0)
			PERROR("Could not requeue old buffer");
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		*ftime = vct->last_time;
		vct->new_index = -1;
		return &vct->bufs[vct->last_index * vct->planes];
	} else if(vct->new_seq <= vct->last_seq)
	{
		/* New frame out of sequence! */
		EPRINT("New frame out of sequence - ignoring\n");
		if(vct->new_index >= 0) {
			vct->cbuf.index = vct->new_index;
			if(ioctl(vct->fd, VIDIOC_QBUF, &vct->cbuf)!=0)
				PERROR("Could not requeue ignored buffer");
		}
		vct->new_index = -1;
		return NULL;
	} else {
		/* New frame - lost something between though */
		vct->last_seq++;
		vct->last_time += vct->frame_period;
		*ftime = vct->last_time;
		return &vct->bufs[vct->last_index * vct->planes];
	}
	return NULL;
}

double v4l2_core_status(struct cap_core_t * cct)
{
	struct v4l2_core_t * vct = (struct v4l2_core_t *)cct->priv;
	double qd;

	if(!vct)
		return 0;
	if(vct->new_index >= 0) {
		qd = vct->new_time;
	} else if(vct->last_index >= 0) {
		qd = vct->last_time;
	} else {
		return 0;
	}
	qd = (double)getntime() - qd;
	qd /= (double)vct->nbufs * (double)vct->frame_period;
	return qd * 100.0;
}

ntime_t v4l2_core_ftime(struct cap_core_t * cct)
{
	struct v4l2_core_t * vct = (struct v4l2_core_t *)cct->priv;

	if(!vct)
		return 0;

	return vct->frame_period;
}

void v4l2_core_usage()
{
	MPRINT("V4L2 capture core usage:\n");
	MPRINT("\t-v <device name> capture device name        [/dev/video]\n");
	MPRINT("\t-w <width> picture width                    [384]\n");
	MPRINT("\t-h <height> picture height                  [288]\n");
	MPRINT("\t-norm <PAL|PAL-M|PAL-N|PAL-60|NTSC|\n");
	MPRINT("\t       NTSC-N|NTSC-44|SECAM> set input norm [unchanged]\n");
	MPRINT("\t-input <input name> set input               [unchanged]\n");
	MPRINT("\t-frequency <MHz> set input frequency        [unchanged]\n");
}

struct cap_core_t v4l2_core;

struct cap_core_t * v4l2_core_init(char ** args, unsigned int **pfmt_p)
{
	struct cap_core_t * cct;
	struct v4l2_core_t * vct;
 	int i, j;
	struct v4l2_format cfmt;
	struct v4l2_requestbuffers crbuf;
	struct v4l2_control ctrl;
	struct v4l2_streamparm cparm;
	char * fdev = NULL;
	int N = -1;
	char * norm = NULL;
	char * inp = NULL;
	double freq = -1;
	unsigned int *pfmt = *pfmt_p;
	NV_ARGS;
	
	cct = malloc(sizeof(struct cap_core_t));
	if(!cct) {
		PERROR("Could no malloc new cap core");
		return NULL;
	}
	vct = malloc(sizeof(struct v4l2_core_t));
	if(!vct) {
		PERROR("Could no malloc new v4l2 core");
		free(cct);
		return NULL;
	}
	memset(vct, 0, sizeof(struct v4l2_core_t));
	memcpy(cct, &v4l2_core, sizeof(struct cap_core_t));
	cct->priv = (void *)vct;
	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			vct->width = atoi(NV_THIS_ARG);
			if(vct->width < 16) {
				EPRINT("Invalid argument to -w (>=16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			vct->height = atoi(NV_THIS_ARG);
			if(vct->height < 16) {
				EPRINT("Invalid argument to -h (>=16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-v") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -v.\n");
				goto usage;
			}
			fdev = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-N") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -N.\n");
				goto usage;
			}
			N = atoi(NV_THIS_ARG);
			if(N < 4) {
				EPRINT("Invalid argument to -N (>=4).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-norm") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -norm.\n");
				goto usage;
			}
			norm = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-input") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -input.\n");
				goto usage;
			}
			inp = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-frequency") == 0) {
			if(!NV_NEXT_ARG) { 
				EPRINT("Missing argument to -frequency.\n");
				goto usage;
			}
			freq = atof(NV_THIS_ARG);
			if(freq < 0) {
				EPRINT("Invalid argument to -frequency (>=0).\n");
				goto usage;
			}
		}
	}
	IPRINT("Size: %dx%d from %s\n", vct->width, vct->height, fdev);

	vct->fd = open(fdev, O_RDWR);
	if(vct->fd<=0) {
		PERROR("Open capture device");
		goto error0;
	}
	
	if(inp) {
		struct v4l2_input input;
		int i = 0;
		
		do {
			memset(&input, 0, sizeof(input));
			input.index = i;
			if(ioctl(vct->fd, VIDIOC_ENUMINPUT, &input)) {
				EPRINT("Unknown input '%s'\n", inp);
				goto error1;
			}
			if(strcmp(input.name, inp) == 0)
				break;
			i++;
		}  while (i > 0);
		if(ioctl(vct->fd, VIDIOC_S_INPUT, &i)) {
			PERROR("Could not set input");
			goto error1;
		}
	}
	
	if(norm) {
		struct v4l2_enumstd std;
		int i = 0;
		do {
			memset(&std, 0, sizeof(std));
			std.index = i;
			if(ioctl(vct->fd, VIDIOC_ENUMSTD, &std)) {
				EPRINT("Unknown norm '%s'\n", norm);
				goto error1;
			}
			if(strcmp(std.std.name, norm) == 0)
				break;
			i++;
		}  while (i > 0);
		if(ioctl(vct->fd, VIDIOC_S_STD, &std.std)) {
			PERROR("Could not set norm");
			goto error1;
		}
	}
	
	if(freq > 0) {
		int i = (int)(freq * 16.0);
		if(ioctl(vct->fd, VIDIOC_S_FREQ, &i)) {
			PERROR("Could not set frequency");
			goto error1;
		}
	}
	
	while(*pfmt != NV_END_FORMAT) { 
		switch(*pfmt) {
			case NV_YUV420:
				IPRINT("Trying format YUV420\n");
				vct->planes = 3;
				vct->pfmt = V4L2_PIX_FMT_YUV420;
				break;
			case NV_RGB24:
				IPRINT("Trying format RGB24\n");
				vct->planes = vct->height;
				vct->pfmt = V4L2_PIX_FMT_BGR24;
				break;
			case NV_YUV422:
				IPRINT("Trying format YUV422\n");
				vct->planes = vct->height;
				vct->pfmt = V4L2_PIX_FMT_YUYV;
				break;
			default:
				IPRINT("Cannot currently handle requested format\n");
				pfmt++;
				continue;
		}

		memset(&cfmt, 0, sizeof(cfmt));
 		cfmt.type = V4L2_BUF_TYPE_CAPTURE;
		cfmt.fmt.pix.width = vct->width;
		cfmt.fmt.pix.height = vct->height;
		cfmt.fmt.pix.pixelformat = vct->pfmt;
		cfmt.fmt.pix.flags = V4L2_FMT_FLAG_INTERLACED;
		if (ioctl(vct->fd, VIDIOC_S_FMT, &cfmt)) {
			PERROR("s_fmt");
			pfmt++;
			continue;
		}
		if(cfmt.fmt.pix.pixelformat != vct->pfmt) {
			IPRINT("Could not get requested pixel format.\n");
			pfmt++;
			continue;
		}
		break;
	}
	if(*pfmt == NV_END_FORMAT) {
		EPRINT("Could not find any matching format!\n");
		goto error1;
	}
	*pfmt_p = pfmt;
	if(cfmt.fmt.pix.width != vct->width) {
		EPRINT("Could not get requested width.\n");
		goto error1;
	}
	if(cfmt.fmt.pix.height != vct->height) {
		EPRINT("Could not get requested height.\n");
		goto error1;
	}

	memset(&cparm, 0, sizeof(cparm));
	cparm.type = V4L2_BUF_TYPE_CAPTURE;
	if (ioctl(vct->fd, VIDIOC_G_PARM, &cparm) != 0) {
		PERROR("Could not request parameters");
		goto error1;
	}
	vct->frame_period = (ntime_t)cparm.parm.capture.timeperframe * 100;
	IPRINT("Frame period %lluns\n", vct->frame_period);

	memset(&crbuf, 0, sizeof(crbuf));
	crbuf.type = V4L2_BUF_TYPE_CAPTURE;
	crbuf.count = N;
	if (ioctl(vct->fd, VIDIOC_REQBUFS, &crbuf) != 0) {
		PERROR("Could not request buffers");
		goto error1;
	}
	if (crbuf.count < 4) {
		EPRINT("Could not get enough buffers (at least 4 required)\n");
		goto error1;
	}
	vct->nbufs = crbuf.count;
	vct->bufs = (unsigned char **)malloc(sizeof(unsigned char *) * vct->nbufs * vct->planes);
	if(!vct->bufs) {
		EPRINT("Could not allocate memory  (bufs)");
		goto error1;
	}
	
	IPRINT("Got %i buffers\n", vct->nbufs);
  
	for(i=0; i<vct->nbufs; i++) {
		vct->cbuf.type = V4L2_BUF_TYPE_CAPTURE;
		vct->cbuf.index = i;
		if(ioctl(vct->fd, VIDIOC_QUERYBUF, &vct->cbuf)!=0) {
			PERROR("Querybuf failed");
			goto error2;
		}
		vct->bufs[i * vct->planes] = (unsigned char *)mmap(NULL, vct->cbuf.length, PROT_READ|PROT_WRITE, MAP_SHARED, vct->fd, vct->cbuf.offset);
		if((int)(vct->bufs[i * vct->planes])==-1) {
			PERROR("MMAP failed");
			goto error2;
		}
		switch(vct->pfmt) {
			case V4L2_PIX_FMT_YUV420:
				vct->bufs[i * vct->planes + 1] =
					vct->bufs[i * vct->planes] +
					vct->width * vct->height;
				vct->bufs[i * vct->planes + 2] =
					vct->bufs[i * vct->planes + 1] +
					(vct->width * vct->height) / 4;
				break;
			case V4L2_PIX_FMT_YUYV:
			case V4L2_PIX_FMT_BGR24:
				for(j = 1; j < vct->height; j++) {
					vct->bufs[i * vct->planes + j] =
						vct->bufs[i * vct->planes + (j - 1)]
						+ cfmt.fmt.pix.bytesperline;
				}
		}
		if(ioctl(vct->fd, VIDIOC_QBUF, &vct->cbuf)!=0) {
			PERROR("Could not queue new buffer");
			goto error2;
		}
	}

	memset(&ctrl, 0, sizeof(ctrl));
	ctrl.id = V4L2_CID_AUDIO_MUTE;  
	if(ioctl(vct->fd, VIDIOC_G_CTRL, &ctrl)!=0) {
		PERROR("Could not get mute status (continuing anyway)");
	}
	vct->old_mute = ctrl.value;
	ctrl.value = 0;
	if(ioctl(vct->fd, VIDIOC_S_CTRL, &ctrl)!=0) {
		PERROR("Could not unmute audio (continuing anyway)");
	}
	
	vct->last_index = -1;
	vct->last_seq = -1;
	vct->new_index = -1;
	vct->new_seq = -1;
	
	i = V4L2_BUF_TYPE_CAPTURE;
	if(ioctl(vct->fd, VIDIOC_STREAMON, &i)!=0) {
		PERROR("Could not streamon");
		goto error2;
	}
	
#if 0
	#define BTTV_DUMPRISC _IOR('v',  BASE_VIDIOCPRIVATE+0, int)
	usleep(50000);
	ioctl(vct->fd, BTTV_DUMPRISC, &i);
#endif

	return cct;
error2:
	free(vct->bufs);
error1:
	close(vct->fd);
error0:
	free(cct);
	free(vct);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(cct);
	free(vct);
	return NULL;
}

struct cap_core_t v4l2_core = {
	cap_core_init: v4l2_core_init,
	cap_core_close: v4l2_core_close,
	cap_core_next: v4l2_core_next,
	cap_core_status: v4l2_core_status,
	cap_core_ftime: v4l2_core_ftime,
	cap_core_overrides: NULL,
	cap_core_usage: v4l2_core_usage,
	priv: NULL
};

#else

struct cap_core_t * v4l2_null_core_init(char ** args, unsigned int **pfmt)
{
	return NULL;
}

void v4l2_null_core_usage()
{
}

struct cap_core_t v4l2_core = {
	cap_core_init: v4l2_null_core_init,
	cap_core_close: NULL,
	cap_core_next: NULL,
	cap_core_status: NULL,
	cap_core_ftime: NULL,
	cap_core_overrides: NULL,
	cap_core_usage: v4l2_null_core_usage,
	priv: NULL
};

#endif
